# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: libp2p/pubsub/pb/rpc.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'libp2p/pubsub/pb/rpc.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1alibp2p/pubsub/pb/rpc.proto\x12\tpubsub.pb\"\xca\x01\n\x03RPC\x12-\n\rsubscriptions\x18\x01 \x03(\x0b\x32\x16.pubsub.pb.RPC.SubOpts\x12#\n\x07publish\x18\x02 \x03(\x0b\x32\x12.pubsub.pb.Message\x12*\n\x07\x63ontrol\x18\x03 \x01(\x0b\x32\x19.pubsub.pb.ControlMessage\x12\x14\n\x0csenderRecord\x18\x04 \x01(\x0c\x1a-\n\x07SubOpts\x12\x11\n\tsubscribe\x18\x01 \x01(\x08\x12\x0f\n\x07topicid\x18\x02 \x01(\t\"i\n\x07Message\x12\x0f\n\x07\x66rom_id\x18\x01 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\r\n\x05seqno\x18\x03 \x01(\x0c\x12\x10\n\x08topicIDs\x18\x04 \x03(\t\x12\x11\n\tsignature\x18\x05 \x01(\x0c\x12\x0b\n\x03key\x18\x06 \x01(\x0c\"\xe0\x01\n\x0e\x43ontrolMessage\x12&\n\x05ihave\x18\x01 \x03(\x0b\x32\x17.pubsub.pb.ControlIHave\x12&\n\x05iwant\x18\x02 \x03(\x0b\x32\x17.pubsub.pb.ControlIWant\x12&\n\x05graft\x18\x03 \x03(\x0b\x32\x17.pubsub.pb.ControlGraft\x12&\n\x05prune\x18\x04 \x03(\x0b\x32\x17.pubsub.pb.ControlPrune\x12.\n\tidontwant\x18\x05 \x03(\x0b\x32\x1b.pubsub.pb.ControlIDontWant\"3\n\x0c\x43ontrolIHave\x12\x0f\n\x07topicID\x18\x01 \x01(\t\x12\x12\n\nmessageIDs\x18\x02 \x03(\t\"\"\n\x0c\x43ontrolIWant\x12\x12\n\nmessageIDs\x18\x01 \x03(\t\"\x1f\n\x0c\x43ontrolGraft\x12\x0f\n\x07topicID\x18\x01 \x01(\t\"T\n\x0c\x43ontrolPrune\x12\x0f\n\x07topicID\x18\x01 \x01(\t\x12\"\n\x05peers\x18\x02 \x03(\x0b\x32\x13.pubsub.pb.PeerInfo\x12\x0f\n\x07\x62\x61\x63koff\x18\x03 \x01(\x04\"&\n\x10\x43ontrolIDontWant\x12\x12\n\nmessageIDs\x18\x01 \x03(\x0c\"4\n\x08PeerInfo\x12\x0e\n\x06peerID\x18\x01 \x01(\x0c\x12\x18\n\x10signedPeerRecord\x18\x02 \x01(\x0c\"\x87\x03\n\x0fTopicDescriptor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x31\n\x04\x61uth\x18\x02 \x01(\x0b\x32#.pubsub.pb.TopicDescriptor.AuthOpts\x12/\n\x03\x65nc\x18\x03 \x01(\x0b\x32\".pubsub.pb.TopicDescriptor.EncOpts\x1a|\n\x08\x41uthOpts\x12:\n\x04mode\x18\x01 \x01(\x0e\x32,.pubsub.pb.TopicDescriptor.AuthOpts.AuthMode\x12\x0c\n\x04keys\x18\x02 \x03(\x0c\"&\n\x08\x41uthMode\x12\x08\n\x04NONE\x10\x00\x12\x07\n\x03KEY\x10\x01\x12\x07\n\x03WOT\x10\x02\x1a\x83\x01\n\x07\x45ncOpts\x12\x38\n\x04mode\x18\x01 \x01(\x0e\x32*.pubsub.pb.TopicDescriptor.EncOpts.EncMode\x12\x11\n\tkeyHashes\x18\x02 \x03(\x0c\"+\n\x07\x45ncMode\x12\x08\n\x04NONE\x10\x00\x12\r\n\tSHAREDKEY\x10\x01\x12\x07\n\x03WOT\x10\x02')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nanolib.pubsub.pb.rpc_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_RPC']._serialized_start=42
  _globals['_RPC']._serialized_end=244
  _globals['_RPC_SUBOPTS']._serialized_start=199
  _globals['_RPC_SUBOPTS']._serialized_end=244
  _globals['_MESSAGE']._serialized_start=246
  _globals['_MESSAGE']._serialized_end=351
  _globals['_CONTROLMESSAGE']._serialized_start=354
  _globals['_CONTROLMESSAGE']._serialized_end=578
  _globals['_CONTROLIHAVE']._serialized_start=580
  _globals['_CONTROLIHAVE']._serialized_end=631
  _globals['_CONTROLIWANT']._serialized_start=633
  _globals['_CONTROLIWANT']._serialized_end=667
  _globals['_CONTROLGRAFT']._serialized_start=669
  _globals['_CONTROLGRAFT']._serialized_end=700
  _globals['_CONTROLPRUNE']._serialized_start=702
  _globals['_CONTROLPRUNE']._serialized_end=786
  _globals['_CONTROLIDONTWANT']._serialized_start=788
  _globals['_CONTROLIDONTWANT']._serialized_end=826
  _globals['_PEERINFO']._serialized_start=828
  _globals['_PEERINFO']._serialized_end=880
  _globals['_TOPICDESCRIPTOR']._serialized_start=883
  _globals['_TOPICDESCRIPTOR']._serialized_end=1274
  _globals['_TOPICDESCRIPTOR_AUTHOPTS']._serialized_start=1016
  _globals['_TOPICDESCRIPTOR_AUTHOPTS']._serialized_end=1140
  _globals['_TOPICDESCRIPTOR_AUTHOPTS_AUTHMODE']._serialized_start=1102
  _globals['_TOPICDESCRIPTOR_AUTHOPTS_AUTHMODE']._serialized_end=1140
  _globals['_TOPICDESCRIPTOR_ENCOPTS']._serialized_start=1143
  _globals['_TOPICDESCRIPTOR_ENCOPTS']._serialized_end=1274
  _globals['_TOPICDESCRIPTOR_ENCOPTS_ENCMODE']._serialized_start=1231
  _globals['_TOPICDESCRIPTOR_ENCOPTS_ENCMODE']._serialized_end=1274
# @@protoc_insertion_point(module_scope)



from collections.abc import (
    AsyncIterator,
)
from types import (
    TracebackType,
)
from typing import (
    Optional,
    Type,
)

import trio

from nanolib.abc import (
    ISubscriptionAPI,
)
from nanolib.custom_types import (
    UnsubscribeFn,
)

from .pb import (
    rpc_pb2,
)


class BaseSubscriptionAPI(ISubscriptionAPI):
    async def __aenter__(self) -> "BaseSubscriptionAPI":
        await trio.lowlevel.checkpoint()
        return self

    async def __aexit__(
        self,
        exc_type: "Optional[Type[BaseException]]",
        exc_value: "Optional[BaseException]",
        traceback: "Optional[TracebackType]",
    ) -> None:
        await self.unsubscribe()


class TrioSubscriptionAPI(BaseSubscriptionAPI):
    receive_channel: "trio.MemoryReceiveChannel[rpc_pb2.Message]"
    unsubscribe_fn: UnsubscribeFn

    def __init__(
        self,
        receive_channel: "trio.MemoryReceiveChannel[rpc_pb2.Message]",
        unsubscribe_fn: UnsubscribeFn,
    ) -> None:
        self.receive_channel = receive_channel
        self.unsubscribe_fn = unsubscribe_fn

    async def unsubscribe(self) -> None:
        await self.unsubscribe_fn()

    def __aiter__(self) -> AsyncIterator[rpc_pb2.Message]:
        return self.receive_channel.__aiter__()

    async def get(self) -> rpc_pb2.Message:
        return await self.receive_channel.receive()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class HopMessage(google.protobuf.message.Message):
    """Circuit v2 message types"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Type:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[HopMessage._Type.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        RESERVE: HopMessage._Type.ValueType  # 0
        CONNECT: HopMessage._Type.ValueType  # 1
        STATUS: HopMessage._Type.ValueType  # 2

    class Type(_Type, metaclass=_TypeEnumTypeWrapper): ...
    RESERVE: HopMessage.Type.ValueType  # 0
    CONNECT: HopMessage.Type.ValueType  # 1
    STATUS: HopMessage.Type.ValueType  # 2

    TYPE_FIELD_NUMBER: builtins.int
    PEER_FIELD_NUMBER: builtins.int
    RESERVATION_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    SENDERRECORD_FIELD_NUMBER: builtins.int
    type: global___HopMessage.Type.ValueType
    peer: builtins.bytes
    senderRecord: builtins.bytes
    """Envelope(PeerRecord)"""
    @property
    def reservation(self) -> global___Reservation: ...
    @property
    def limit(self) -> global___Limit: ...
    @property
    def status(self) -> global___Status: ...
    def __init__(
        self,
        *,
        type: global___HopMessage.Type.ValueType = ...,
        peer: builtins.bytes = ...,
        reservation: global___Reservation | None = ...,
        limit: global___Limit | None = ...,
        status: global___Status | None = ...,
        senderRecord: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_senderRecord", b"_senderRecord", "limit", b"limit", "reservation", b"reservation", "senderRecord", b"senderRecord", "status", b"status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_senderRecord", b"_senderRecord", "limit", b"limit", "peer", b"peer", "reservation", b"reservation", "senderRecord", b"senderRecord", "status", b"status", "type", b"type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["_senderRecord", b"_senderRecord"]) -> typing.Literal["senderRecord"] | None: ...

global___HopMessage = HopMessage

@typing.final
class StopMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Type:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[StopMessage._Type.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        CONNECT: StopMessage._Type.ValueType  # 0
        STATUS: StopMessage._Type.ValueType  # 1

    class Type(_Type, metaclass=_TypeEnumTypeWrapper): ...
    CONNECT: StopMessage.Type.ValueType  # 0
    STATUS: StopMessage.Type.ValueType  # 1

    TYPE_FIELD_NUMBER: builtins.int
    PEER_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    SENDERRECORD_FIELD_NUMBER: builtins.int
    type: global___StopMessage.Type.ValueType
    peer: builtins.bytes
    senderRecord: builtins.bytes
    """Envelope(PeerRecord) encoded"""
    @property
    def status(self) -> global___Status: ...
    def __init__(
        self,
        *,
        type: global___StopMessage.Type.ValueType = ...,
        peer: builtins.bytes = ...,
        status: global___Status | None = ...,
        senderRecord: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_senderRecord", b"_senderRecord", "senderRecord", b"senderRecord", "status", b"status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_senderRecord", b"_senderRecord", "peer", b"peer", "senderRecord", b"senderRecord", "status", b"status", "type", b"type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["_senderRecord", b"_senderRecord"]) -> typing.Literal["senderRecord"] | None: ...

global___StopMessage = StopMessage

@typing.final
class Reservation(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VOUCHER_FIELD_NUMBER: builtins.int
    SIGNATURE_FIELD_NUMBER: builtins.int
    EXPIRE_FIELD_NUMBER: builtins.int
    voucher: builtins.bytes
    signature: builtins.bytes
    expire: builtins.int
    def __init__(
        self,
        *,
        voucher: builtins.bytes = ...,
        signature: builtins.bytes = ...,
        expire: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["expire", b"expire", "signature", b"signature", "voucher", b"voucher"]) -> None: ...

global___Reservation = Reservation

@typing.final
class Limit(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DURATION_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    duration: builtins.int
    data: builtins.int
    def __init__(
        self,
        *,
        duration: builtins.int = ...,
        data: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["data", b"data", "duration", b"duration"]) -> None: ...

global___Limit = Limit

@typing.final
class Status(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Code:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _CodeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Status._Code.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        OK: Status._Code.ValueType  # 0
        RESERVATION_REFUSED: Status._Code.ValueType  # 100
        RESOURCE_LIMIT_EXCEEDED: Status._Code.ValueType  # 101
        PERMISSION_DENIED: Status._Code.ValueType  # 102
        CONNECTION_FAILED: Status._Code.ValueType  # 200
        DIAL_REFUSED: Status._Code.ValueType  # 201
        STOP_FAILED: Status._Code.ValueType  # 300
        MALFORMED_MESSAGE: Status._Code.ValueType  # 400

    class Code(_Code, metaclass=_CodeEnumTypeWrapper): ...
    OK: Status.Code.ValueType  # 0
    RESERVATION_REFUSED: Status.Code.ValueType  # 100
    RESOURCE_LIMIT_EXCEEDED: Status.Code.ValueType  # 101
    PERMISSION_DENIED: Status.Code.ValueType  # 102
    CONNECTION_FAILED: Status.Code.ValueType  # 200
    DIAL_REFUSED: Status.Code.ValueType  # 201
    STOP_FAILED: Status.Code.ValueType  # 300
    MALFORMED_MESSAGE: Status.Code.ValueType  # 400

    CODE_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    code: global___Status.Code.ValueType
    message: builtins.str
    def __init__(
        self,
        *,
        code: global___Status.Code.ValueType = ...,
        message: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["code", b"code", "message", b"message"]) -> None: ...

global___Status = Status



import logging

from multiaddr import (
    Multiaddr,
)
import trio

from nanolib.abc import (
    IHost,
    INetStream,
    IPeerStore,
)
from nanolib.crypto.serialization import (
    deserialize_public_key,
)
from nanolib.custom_types import (
    StreamHandlerFn,
    TProtocol,
)
from nanolib.network.stream.exceptions import (
    StreamClosed,
)
from nanolib.peer.envelope import consume_envelope
from nanolib.peer.id import (
    ID,
)
from nanolib.utils import (
    get_agent_version,
    varint,
)
from nanolib.utils.varint import (
    read_length_prefixed_protobuf,
)

from ..identify.identify import (
    _mk_identify_protobuf,
)
from ..identify.pb.identify_pb2 import (
    Identify,
)

logger = logging.getLogger(__name__)

# Protocol ID for identify/push
ID_PUSH = TProtocol("/ipfs/id/push/1.0.0")
PROTOCOL_VERSION = "ipfs/0.1.0"
AGENT_VERSION = get_agent_version()
CONCURRENCY_LIMIT = 10


def identify_push_handler_for(
    host: IHost, use_varint_format: bool = True
) -> StreamHandlerFn:
    """
    Create a handler for the identify/push protocol.

    This handler receives pushed identify messages from remote peers and updates
    the local peerstore with the new information.

    Args:
        host: The libp2p host.
        use_varint_format: True=length-prefixed, False=raw protobuf.

    """

    async def handle_identify_push(stream: INetStream) -> None:
        peer_id = stream.muxed_conn.peer_id

        try:
            # Use the utility function to read the protobuf message
            data = await read_length_prefixed_protobuf(stream, use_varint_format)

            identify_msg = Identify()
            identify_msg.ParseFromString(data)

            # Update the peerstore with the new information
            await _update_peerstore_from_identify(
                host.get_peerstore(), peer_id, identify_msg
            )

            logger.debug("Successfully processed identify/push from peer %s", peer_id)

            # Send acknowledgment to indicate successful processing
            # This ensures the sender knows the message was received before closing
            await stream.write(b"OK")

        except StreamClosed:
            logger.debug(
                "Stream closed while processing identify/push from %s", peer_id
            )
        except Exception as e:
            logger.error("Error processing identify/push from %s: %s", peer_id, e)
        finally:
            # Close the stream after processing
            try:
                await stream.close()
            except Exception:
                pass  # Ignore errors when closing

    return handle_identify_push


async def _update_peerstore_from_identify(
    peerstore: IPeerStore, peer_id: ID, identify_msg: Identify
) -> None:
    """
    Update the peerstore with information from an identify message.

    This function handles partial updates, where only some fields may be present
    in the identify message.

    Security: Signed peer records are validated to ensure the peer ID in the
    record matches the sender's peer ID to prevent peer ID spoofing attacks.
    """
    # Update public key if present
    if identify_msg.HasField("public_key"):
        try:
            peerstore.add_protocols(peer_id, [])
            pubkey = deserialize_public_key(identify_msg.public_key)
            peerstore.add_pubkey(peer_id, pubkey)
        except Exception as e:
            logger.error("Error updating public key for peer %s: %s", peer_id, e)

    # Update listen addresses if present
    if identify_msg.listen_addrs:
        try:
            addrs = [Multiaddr(addr) for addr in identify_msg.listen_addrs]
            for addr in addrs:
                peerstore.add_addr(peer_id, addr, 7200)  # 2 hours TTL
        except Exception as e:
            logger.error("Error updating listen addresses for peer %s: %s", peer_id, e)

    # Update protocols if present
    if identify_msg.protocols:
        try:
            peerstore.add_protocols(peer_id, identify_msg.protocols)
        except Exception as e:
            logger.error("Error updating protocols for peer %s: %s", peer_id, e)

    # Update from signed peer record if present
    if identify_msg.HasField("signedPeerRecord"):
        try:
            envelope, record = consume_envelope(
                identify_msg.signedPeerRecord, "libp2p-peer-record"
            )
            # Cross-check peer-id consistency
            # Security: Reject signed peer records where the peer ID doesn't match
            # the sender's peer ID to prevent peer ID spoofing attacks
            if record.peer_id != peer_id:
                logger.warning(
                    "SignedPeerRecord peer-id mismatch: record=%s, sender=%s. "
                    "Ignoring.",
                    record.peer_id,
                    peer_id,
                )
                return  # Reject forged record - peer ID mismatch

            if not peerstore.consume_peer_record(envelope, 7200):
                logger.error(
                    "Updating Certified-Addr-Book was unsuccessful for %s", peer_id
                )
        except Exception as e:
            logger.error(
                "Error updating the certified addr book for peer %s: %s", peer_id, e
            )

    # Update observed address if present
    if identify_msg.HasField("observed_addr") and identify_msg.observed_addr:
        try:
            observed_addr = Multiaddr(identify_msg.observed_addr)
            peerstore.add_addr(peer_id, observed_addr, 7200)
        except Exception as e:
            logger.error("Error updating observed address for peer %s: %s", peer_id, e)


async def push_identify_to_peer(
    host: IHost,
    peer_id: ID,
    observed_multiaddr: Multiaddr | None = None,
    limit: trio.Semaphore = trio.Semaphore(CONCURRENCY_LIMIT),
    use_varint_format: bool = True,
) -> bool:
    """
    Push an identify message to a specific peer.

    This function opens a stream to the peer using the identify/push protocol,
    sends the identify message, and closes the stream.

    Args:
        host: The libp2p host.
        peer_id: The peer ID to push to.
        observed_multiaddr: The observed multiaddress (optional).
        limit: Semaphore for concurrency control.
        use_varint_format: True=length-prefixed, False=raw protobuf.

    Returns:
        bool: True if the push was successful, False otherwise.

    """
    async with limit:
        try:
            # Create a new stream to the peer using the identify/push protocol
            stream = await host.new_stream(peer_id, [ID_PUSH])

            # Create the identify message
            identify_msg = _mk_identify_protobuf(host, observed_multiaddr)
            response = identify_msg.SerializeToString()

            if use_varint_format:
                # Send length-prefixed identify message
                await stream.write(varint.encode_uvarint(len(response)))
                await stream.write(response)
            else:
                # Send raw protobuf message
                await stream.write(response)

            # Wait for acknowledgment from the receiver with timeout
            # This ensures the message was processed before closing
            try:
                with trio.move_on_after(1.0):  # 1 second timeout
                    ack = await stream.read(2)  # Read "OK" acknowledgment
                    if ack != b"OK":
                        logger.warning(
                            "Unexpected acknowledgment from peer %s: %s", peer_id, ack
                        )
            except Exception as e:
                logger.debug("No acknowledgment received from peer %s: %s", peer_id, e)
                # Continue anyway, as the message might have been processed

            # Close the stream after acknowledgment (or timeout)
            await stream.close()

            logger.debug("Successfully pushed identify to peer %s", peer_id)
            return True
        except Exception as e:
            logger.error("Error pushing identify to peer %s: %s", peer_id, e)
            return False


async def push_identify_to_peers(
    host: IHost,
    peer_ids: set[ID] | None = None,
    observed_multiaddr: Multiaddr | None = None,
    use_varint_format: bool = True,
) -> None:
    """
    Push an identify message to multiple peers in parallel.

    If peer_ids is None, push to all connected peers.

    Args:
        host: The libp2p host.
        peer_ids: Set of peer IDs to push to (if None, push to all connected peers).
        observed_multiaddr: The observed multiaddress (optional).
        use_varint_format: True=length-prefixed, False=raw protobuf.

    """
    if peer_ids is None:
        # Get all connected peers
        peer_ids = set(host.get_connected_peers())

    # Create a single shared semaphore for concurrency control
    limit = trio.Semaphore(CONCURRENCY_LIMIT)

    # Push to each peer in parallel using a trio.Nursery
    # limiting concurrent connections to CONCURRENCY_LIMIT
    async with trio.open_nursery() as nursery:
        for peer_id in peer_ids:
            nursery.start_soon(
                push_identify_to_peer,
                host,
                peer_id,
                observed_multiaddr,
                limit,
                use_varint_format,
            )



from .identify_push import (
    ID_PUSH,
    identify_push_handler_for,
    push_identify_to_peer,
    push_identify_to_peers,
)

__all__ = [
    "ID_PUSH",
    "identify_push_handler_for",
    "push_identify_to_peer",
    "push_identify_to_peers",
]



"""
Protocol rate limiter for resource management.

This module implements protocol-specific rate limiting that provides
fine-grained control over protocol usage for production-ready
resource management.
"""

from __future__ import annotations

from dataclasses import dataclass
import time
from typing import Any

from nanolib.peer.id import ID

from .rate_limiter import (
    RateLimitConfig,
    RateLimiter,
    RateLimitScope,
    RateLimitStrategy,
)


@dataclass
class ProtocolRateLimitConfig:
    """Configuration for protocol rate limiter."""

    # Protocol identification
    protocol_name: str
    protocol_version: str | None = None

    # Rate limiting parameters
    refill_rate: float = 10.0  # Tokens per second
    capacity: float = 100.0  # Maximum tokens in bucket
    initial_tokens: float = 0.0  # Initial tokens in bucket

    # Burst handling
    allow_burst: bool = True  # Allow burst up to capacity
    burst_multiplier: float = 1.0  # Burst capacity multiplier

    # Time parameters
    time_window_seconds: float = 1.0  # Time window for rate calculation
    min_interval_seconds: float = 0.001  # Minimum interval between refills

    # Scope parameters
    scope: RateLimitScope = RateLimitScope.PER_PEER  # Rate limiting scope
    max_peers: int = 1000  # Maximum number of peers to track
    max_connections: int = 1000  # Maximum number of connections to track

    # Monitoring parameters
    enable_monitoring: bool = True  # Enable monitoring and statistics
    max_history_size: int = 1000  # Maximum history size for monitoring

    # Protocol-specific parameters
    max_concurrent_requests: int = 10  # Maximum concurrent requests
    request_timeout_seconds: float = 30.0  # Request timeout in seconds
    backoff_factor: float = 2.0  # Backoff factor for retries

    def __post_init__(self) -> None:
        """Validate configuration parameters."""
        if not self.protocol_name:
            raise ValueError("protocol_name cannot be empty")
        if self.refill_rate <= 0:
            raise ValueError("refill_rate must be positive")
        if self.capacity <= 0:
            raise ValueError("capacity must be positive")
        if self.initial_tokens < 0:
            raise ValueError("initial_tokens must be non-negative")
        if self.initial_tokens > self.capacity:
            raise ValueError("initial_tokens cannot exceed capacity")
        if self.burst_multiplier < 1.0:
            raise ValueError("burst_multiplier must be >= 1.0")
        if self.time_window_seconds <= 0:
            raise ValueError("time_window_seconds must be positive")
        if self.min_interval_seconds <= 0:
            raise ValueError("min_interval_seconds must be positive")
        if self.max_history_size <= 0:
            raise ValueError("max_history_size must be positive")
        if self.max_peers <= 0:
            raise ValueError("max_peers must be positive")
        if self.max_connections <= 0:
            raise ValueError("max_connections must be positive")
        if self.max_concurrent_requests <= 0:
            raise ValueError("max_concurrent_requests must be positive")
        if self.request_timeout_seconds <= 0:
            raise ValueError("request_timeout_seconds must be positive")
        if self.backoff_factor < 1.0:
            raise ValueError("backoff_factor must be >= 1.0")


@dataclass
class ProtocolRateLimitStats:
    """Statistics for protocol rate limiter."""

    # Protocol information
    protocol_name: str
    protocol_version: str | None

    # Rate limiting statistics
    current_rate: float
    allowed_requests: int
    denied_requests: int
    total_requests: int

    # Protocol-specific statistics
    concurrent_requests: int
    max_concurrent_requests: int
    request_timeouts: int
    backoff_events: int

    # Scope information
    scope: RateLimitScope
    tracked_entities: int

    # Time information
    time_since_last_request: float
    next_available_time: float

    def to_dict(self) -> dict[str, Any]:
        """Convert statistics to dictionary."""
        return {
            "protocol_name": self.protocol_name,
            "protocol_version": self.protocol_version,
            "current_rate": self.current_rate,
            "allowed_requests": self.allowed_requests,
            "denied_requests": self.denied_requests,
            "total_requests": self.total_requests,
            "concurrent_requests": self.concurrent_requests,
            "max_concurrent_requests": self.max_concurrent_requests,
            "request_timeouts": self.request_timeouts,
            "backoff_events": self.backoff_events,
            "scope": self.scope.value,
            "tracked_entities": self.tracked_entities,
            "time_since_last_request": self.time_since_last_request,
            "next_available_time": self.next_available_time,
        }


class ProtocolRateLimiter:
    """
    Protocol-specific rate limiter implementation.

    This class implements protocol-specific rate limiting that provides
    fine-grained control over protocol usage with support for
    concurrent request tracking and backoff handling.
    """

    def __init__(self, config: ProtocolRateLimitConfig):
        """
        Initialize protocol rate limiter.

        Args:
            config: Protocol rate limiter configuration

        """
        self.config = config

        # Rate limiting state
        self._rate_limiter: RateLimiter = self._create_rate_limiter()

        # Protocol-specific state
        self._concurrent_requests: dict[str, int] = {}
        self._request_start_times: dict[str, list[float]] = {}
        self._backoff_until: dict[str, float] = {}

        # Statistics
        self._request_timeouts: int = 0
        self._backoff_events: int = 0

    def _create_rate_limiter(self) -> RateLimiter:
        """Create the underlying rate limiter."""
        rate_config = RateLimitConfig(
            strategy=RateLimitStrategy.TOKEN_BUCKET,
            scope=self.config.scope,
            refill_rate=self.config.refill_rate,
            capacity=self.config.capacity,
            initial_tokens=self.config.initial_tokens,
            allow_burst=self.config.allow_burst,
            burst_multiplier=self.config.burst_multiplier,
            time_window_seconds=self.config.time_window_seconds,
            min_interval_seconds=self.config.min_interval_seconds,
            enable_monitoring=self.config.enable_monitoring,
            max_history_size=self.config.max_history_size,
            max_peers=self.config.max_peers,
            max_connections=self.config.max_connections,
        )
        return RateLimiter(rate_config)

    def _get_entity_key(
        self,
        peer_id: ID | None = None,
        connection_id: str | None = None,
    ) -> str:
        """
        Get entity key for rate limiting.

        Args:
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)

        Returns:
            Entity key for rate limiting

        """
        if self.config.scope == RateLimitScope.GLOBAL:
            return "global"
        elif self.config.scope == RateLimitScope.PER_PEER:
            return f"peer:{peer_id}" if peer_id else "unknown"
        elif self.config.scope == RateLimitScope.PER_CONNECTION:
            return f"conn:{connection_id}" if connection_id else "unknown"
        else:
            return "unknown"

    def _cleanup_old_requests(self, entity_key: str, current_time: float) -> None:
        """
        Clean up old requests that have timed out.

        Args:
            entity_key: Entity key for rate limiting
            current_time: Current timestamp

        """
        if entity_key not in self._request_start_times:
            return

        # Remove requests that have timed out
        timeout_threshold = current_time - self.config.request_timeout_seconds
        self._request_start_times[entity_key] = [
            start_time
            for start_time in self._request_start_times[entity_key]
            if start_time > timeout_threshold
        ]

        # Update concurrent request count
        self._concurrent_requests[entity_key] = len(
            self._request_start_times[entity_key]
        )

        # Count timeouts
        if len(self._request_start_times[entity_key]) < self._concurrent_requests.get(
            entity_key, 0
        ):
            self._request_timeouts += 1

    def _is_in_backoff(self, entity_key: str, current_time: float) -> bool:
        """
        Check if entity is in backoff period.

        Args:
            entity_key: Entity key for rate limiting
            current_time: Current timestamp

        Returns:
            True if entity is in backoff period

        """
        if entity_key not in self._backoff_until:
            return False

        if current_time < self._backoff_until[entity_key]:
            return True

        # Backoff period has ended
        del self._backoff_until[entity_key]
        return False

    def _start_backoff(self, entity_key: str, current_time: float) -> None:
        """
        Start backoff period for entity.

        Args:
            entity_key: Entity key for rate limiting
            current_time: Current timestamp

        """
        # Calculate backoff duration
        backoff_duration = (
            self.config.request_timeout_seconds * self.config.backoff_factor
        )

        # Set backoff end time
        self._backoff_until[entity_key] = current_time + backoff_duration

        # Record backoff event
        self._backoff_events += 1

    def try_allow_request(
        self,
        tokens: float = 1.0,
        peer_id: ID | None = None,
        connection_id: str | None = None,
        current_time: float | None = None,
    ) -> bool:
        """
        Try to allow a protocol request based on rate limiting.

        Args:
            tokens: Number of tokens to consume
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)
            current_time: Current timestamp (optional)

        Returns:
            True if request is allowed, False otherwise

        """
        if current_time is None:
            current_time = time.time()

        # Get entity key
        entity_key = self._get_entity_key(peer_id, connection_id)

        # Clean up old requests
        self._cleanup_old_requests(entity_key, current_time)

        # Check if entity is in backoff period
        if self._is_in_backoff(entity_key, current_time):
            return False

        # Check concurrent request limit
        current_concurrent = self._concurrent_requests.get(entity_key, 0)
        if current_concurrent >= self.config.max_concurrent_requests:
            # Start backoff period
            self._start_backoff(entity_key, current_time)
            return False

        # Check rate limiting
        if not self._rate_limiter.try_allow(
            tokens,
            peer_id,
            connection_id,
            self.config.protocol_name,
            None,
            current_time,
        ):
            return False

        # Allow request
        if entity_key not in self._request_start_times:
            self._request_start_times[entity_key] = []

        self._request_start_times[entity_key].append(current_time)
        self._concurrent_requests[entity_key] = len(
            self._request_start_times[entity_key]
        )

        return True

    def allow_request(
        self,
        tokens: float = 1.0,
        peer_id: ID | None = None,
        connection_id: str | None = None,
        current_time: float | None = None,
    ) -> None:
        """
        Allow a protocol request based on rate limiting (raises exception if denied).

        Args:
            tokens: Number of tokens to consume
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)
            current_time: Current timestamp (optional)

        Raises:
            ValueError: If request is denied by rate limiting

        """
        if not self.try_allow_request(tokens, peer_id, connection_id, current_time):
            raise ValueError("Protocol request denied by rate limiting")

    def finish_request(
        self,
        peer_id: ID | None = None,
        connection_id: str | None = None,
        current_time: float | None = None,
    ) -> None:
        """
        Finish a protocol request.

        Args:
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)
            current_time: Current timestamp (optional)

        """
        if current_time is None:
            current_time = time.time()

        # Get entity key
        entity_key = self._get_entity_key(peer_id, connection_id)

        # Remove oldest request
        if (
            entity_key in self._request_start_times
            and self._request_start_times[entity_key]
        ):
            self._request_start_times[entity_key].pop(0)
            self._concurrent_requests[entity_key] = len(
                self._request_start_times[entity_key]
            )

    def get_stats(
        self,
        peer_id: ID | None = None,
        connection_id: str | None = None,
        current_time: float | None = None,
    ) -> ProtocolRateLimitStats:
        """
        Get current statistics for the protocol rate limiter.

        Args:
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)
            current_time: Current timestamp (optional)

        Returns:
            ProtocolRateLimitStats: Current statistics

        """
        if current_time is None:
            current_time = time.time()

        # Get entity key
        entity_key = self._get_entity_key(peer_id, connection_id)

        # Clean up old requests
        self._cleanup_old_requests(entity_key, current_time)

        # Get rate limiter statistics
        rate_stats = self._rate_limiter.get_stats(
            peer_id, connection_id, self.config.protocol_name, None, current_time
        )

        # Get protocol-specific statistics
        concurrent_requests = self._concurrent_requests.get(entity_key, 0)

        # Calculate time since last request
        time_since_last_request = 0.0
        if (
            entity_key in self._request_start_times
            and self._request_start_times[entity_key]
        ):
            last_request_time = max(self._request_start_times[entity_key])
            time_since_last_request = current_time - last_request_time

        return ProtocolRateLimitStats(
            protocol_name=self.config.protocol_name,
            protocol_version=self.config.protocol_version,
            current_rate=rate_stats.current_rate,
            allowed_requests=rate_stats.allowed_requests,
            denied_requests=rate_stats.denied_requests,
            total_requests=rate_stats.total_requests,
            concurrent_requests=concurrent_requests,
            max_concurrent_requests=self.config.max_concurrent_requests,
            request_timeouts=self._request_timeouts,
            backoff_events=self._backoff_events,
            scope=self.config.scope,
            tracked_entities=rate_stats.tracked_entities,
            time_since_last_request=time_since_last_request,
            next_available_time=rate_stats.next_available_time,
        )

    def get_entity_stats(
        self,
        peer_id: ID | None = None,
        connection_id: str | None = None,
    ) -> dict[str, Any]:
        """
        Get entity-specific statistics.

        Args:
            peer_id: Peer ID (optional)
            connection_id: Connection ID (optional)

        Returns:
            Dictionary with entity-specific statistics

        """
        entity_key = self._get_entity_key(peer_id, connection_id)

        # Get rate limiter entity stats
        rate_stats = self._rate_limiter.get_entity_stats(peer_id, connection_id)

        # Add protocol-specific stats
        protocol_stats = {
            "protocol_name": self.config.protocol_name,
            "protocol_version": self.config.protocol_version,
            "concurrent_requests": self._concurrent_requests.get(entity_key, 0),
            "max_concurrent_requests": self.config.max_concurrent_requests,
            "request_start_times": self._request_start_times.get(entity_key, []),
            "in_backoff": entity_key in self._backoff_until,
            "backoff_until": self._backoff_until.get(entity_key, 0.0),
        }

        return {**rate_stats, **protocol_stats}

    def reset(self) -> None:
        """Reset the protocol rate limiter to initial state."""
        self._rate_limiter.reset()
        self._concurrent_requests.clear()
        self._request_start_times.clear()
        self._backoff_until.clear()
        self._request_timeouts = 0
        self._backoff_events = 0

    def __str__(self) -> str:
        """String representation of protocol rate limiter."""
        return (
            f"ProtocolRateLimiter(protocol={self.config.protocol_name}, "
            f"version={self.config.protocol_version}, "
            f"rate={self.config.refill_rate:.2f}/s, "
            f"capacity={self.config.capacity:.2f})"
        )


# Factory functions
def create_protocol_rate_limiter(
    protocol_name: str,
    protocol_version: str | None = None,
    refill_rate: float = 10.0,
    capacity: float = 100.0,
    initial_tokens: float = 0.0,
    allow_burst: bool = True,
    burst_multiplier: float = 1.0,
    time_window_seconds: float = 1.0,
    min_interval_seconds: float = 0.001,
    scope: RateLimitScope = RateLimitScope.PER_PEER,
    max_peers: int = 1000,
    max_connections: int = 1000,
    enable_monitoring: bool = True,
    max_history_size: int = 1000,
    max_concurrent_requests: int = 10,
    request_timeout_seconds: float = 30.0,
    backoff_factor: float = 2.0,
) -> ProtocolRateLimiter:
    """
    Create a protocol rate limiter.

    Args:
        protocol_name: Name of the protocol
        protocol_version: Version of the protocol (optional)
        refill_rate: Tokens per second
        capacity: Maximum tokens in bucket
        initial_tokens: Initial tokens in bucket
        allow_burst: Allow burst up to capacity
        burst_multiplier: Burst capacity multiplier
        time_window_seconds: Time window for rate calculation
        min_interval_seconds: Minimum interval between refills
        scope: Rate limiting scope
        max_peers: Maximum number of peers to track
        max_connections: Maximum number of connections to track
        enable_monitoring: Enable monitoring and statistics
        max_history_size: Maximum history size for monitoring
        max_concurrent_requests: Maximum concurrent requests
        request_timeout_seconds: Request timeout in seconds
        backoff_factor: Backoff factor for retries

    Returns:
        ProtocolRateLimiter: Configured protocol rate limiter

    """
    config = ProtocolRateLimitConfig(
        protocol_name=protocol_name,
        protocol_version=protocol_version,
        refill_rate=refill_rate,
        capacity=capacity,
        initial_tokens=initial_tokens,
        allow_burst=allow_burst,
        burst_multiplier=burst_multiplier,
        time_window_seconds=time_window_seconds,
        min_interval_seconds=min_interval_seconds,
        scope=scope,
        max_peers=max_peers,
        max_connections=max_connections,
        enable_monitoring=enable_monitoring,
        max_history_size=max_history_size,
        max_concurrent_requests=max_concurrent_requests,
        request_timeout_seconds=request_timeout_seconds,
        backoff_factor=backoff_factor,
    )
    return ProtocolRateLimiter(config)


def create_strict_protocol_rate_limiter(
    protocol_name: str,
    protocol_version: str | None = None,
    refill_rate: float = 10.0,
    capacity: float = 100.0,
    initial_tokens: float = 0.0,
) -> ProtocolRateLimiter:
    """
    Create a strict protocol rate limiter (no burst).

    Args:
        protocol_name: Name of the protocol
        protocol_version: Version of the protocol (optional)
        refill_rate: Tokens per second
        capacity: Maximum tokens in bucket
        initial_tokens: Initial tokens in bucket

    Returns:
        ProtocolRateLimiter: Configured strict protocol rate limiter

    """
    return create_protocol_rate_limiter(
        protocol_name=protocol_name,
        protocol_version=protocol_version,
        refill_rate=refill_rate,
        capacity=capacity,
        initial_tokens=initial_tokens,
        allow_burst=False,
        burst_multiplier=1.0,
    )


def create_burst_protocol_rate_limiter(
    protocol_name: str,
    protocol_version: str | None = None,
    refill_rate: float = 10.0,
    capacity: float = 100.0,
    initial_tokens: float = 0.0,
    burst_multiplier: float = 2.0,
) -> ProtocolRateLimiter:
    """
    Create a burst-capable protocol rate limiter.

    Args:
        protocol_name: Name of the protocol
        protocol_version: Version of the protocol (optional)
        refill_rate: Tokens per second
        capacity: Maximum tokens in bucket
        initial_tokens: Initial tokens in bucket
        burst_multiplier: Burst capacity multiplier

    Returns:
        ProtocolRateLimiter: Configured burst-capable protocol rate limiter

    """
    return create_protocol_rate_limiter(
        protocol_name=protocol_name,
        protocol_version=protocol_version,
        refill_rate=refill_rate,
        capacity=capacity,
        initial_tokens=initial_tokens,
        allow_burst=True,
        burst_multiplier=burst_multiplier,
    )



"""Rekey support for Noise protocol sessions."""

from abc import ABC, abstractmethod
import time
from typing import Protocol, runtime_checkable


@runtime_checkable
class RekeyPolicy(Protocol):
    """Protocol for rekey policies."""

    def should_rekey(self, bytes_processed: int, time_elapsed: float) -> bool:
        """
        Determine if a rekey should be performed.

        Args:
            bytes_processed: Number of bytes processed since last rekey
            time_elapsed: Time elapsed since last rekey in seconds

        Returns:
            bool: True if rekey should be performed

        """
        ...


class TimeBasedRekeyPolicy(RekeyPolicy):
    """Rekey policy based on time elapsed."""

    def __init__(self, max_time_seconds: int = 3600):  # 1 hour default
        """
        Initialize with maximum time between rekeys.

        Args:
            max_time_seconds: Maximum time between rekeys in seconds

        """
        self.max_time_seconds = max_time_seconds

    def should_rekey(self, bytes_processed: int, time_elapsed: float) -> bool:
        """
        Check if rekey should be performed based on time.

        Args:
            bytes_processed: Number of bytes processed (ignored for time-based policy)
            time_elapsed: Time elapsed since last rekey in seconds

        Returns:
            bool: True if maximum time has been exceeded

        """
        return time_elapsed >= self.max_time_seconds


class ByteCountRekeyPolicy(RekeyPolicy):
    """Rekey policy based on bytes processed."""

    def __init__(self, max_bytes: int = 1024 * 1024 * 1024):  # 1GB default
        """
        Initialize with maximum bytes between rekeys.

        Args:
            max_bytes: Maximum bytes processed between rekeys

        """
        self.max_bytes = max_bytes

    def should_rekey(self, bytes_processed: int, time_elapsed: float) -> bool:
        """
        Check if rekey should be performed based on bytes processed.

        Args:
            bytes_processed: Number of bytes processed since last rekey
            time_elapsed: Time elapsed (ignored for byte-based policy)

        Returns:
            bool: True if maximum bytes have been processed

        """
        return bytes_processed >= self.max_bytes


class CompositeRekeyPolicy(RekeyPolicy):
    """Rekey policy that combines multiple policies."""

    def __init__(self, policies: list[RekeyPolicy]):
        """
        Initialize with a list of rekey policies.

        Args:
            policies: List of rekey policies to combine

        """
        self.policies = policies

    def should_rekey(self, bytes_processed: int, time_elapsed: float) -> bool:
        """
        Check if rekey should be performed based on any policy.

        Args:
            bytes_processed: Number of bytes processed since last rekey
            time_elapsed: Time elapsed since last rekey in seconds

        Returns:
            bool: True if any policy indicates rekey should be performed

        """
        return any(
            policy.should_rekey(bytes_processed, time_elapsed)
            for policy in self.policies
        )


class RekeyManager:
    """Manager for rekey operations in Noise sessions."""

    def __init__(self, policy: RekeyPolicy | None = None):
        """
        Initialize with an optional rekey policy.

        Args:
            policy: Rekey policy to use

        """
        self.policy = policy or CompositeRekeyPolicy(
            [
                TimeBasedRekeyPolicy(max_time_seconds=3600),  # 1 hour
                ByteCountRekeyPolicy(max_bytes=1024 * 1024 * 1024),  # 1GB
            ]
        )

        self._last_rekey_time = time.time()
        self._bytes_since_rekey = 0
        self._rekey_count = 0

    def update_bytes_processed(self, bytes_count: int) -> None:
        """
        Update the count of bytes processed since last rekey.

        Args:
            bytes_count: Number of bytes processed

        """
        self._bytes_since_rekey += bytes_count

    def should_rekey(self) -> bool:
        """
        Check if a rekey should be performed.

        Returns:
            bool: True if rekey should be performed

        """
        time_elapsed = time.time() - self._last_rekey_time
        return self.policy.should_rekey(self._bytes_since_rekey, time_elapsed)

    def perform_rekey(self) -> None:
        """Mark that a rekey has been performed."""
        self._last_rekey_time = time.time()
        self._bytes_since_rekey = 0
        self._rekey_count += 1

    def get_stats(self) -> dict[str, int | float]:
        """
        Get rekey statistics.

        Returns:
            dict: Statistics about rekey operations

        """
        time_elapsed = time.time() - self._last_rekey_time
        return {
            "rekey_count": self._rekey_count,
            "bytes_since_rekey": self._bytes_since_rekey,
            "time_since_rekey": time_elapsed,
            "last_rekey_time": self._last_rekey_time,
        }

    def reset_stats(self) -> None:
        """Reset rekey statistics."""
        self._last_rekey_time = time.time()
        self._bytes_since_rekey = 0
        self._rekey_count = 0


class RekeyableSession(ABC):
    """Abstract base class for sessions that support rekeying."""

    @abstractmethod
    async def rekey(self) -> None:
        """
        Perform a rekey operation.

        Raises:
            Exception: If rekey operation fails

        """
        pass


class RekeyHandler:
    """Handler for managing rekey operations in Noise sessions."""

    def __init__(
        self, session: RekeyableSession, rekey_manager: RekeyManager | None = None
    ):
        """
        Initialize with a rekeyable session and optional rekey manager.

        Args:
            session: Session that supports rekeying
            rekey_manager: Rekey manager to use

        """
        self.session = session
        self.rekey_manager = rekey_manager or RekeyManager()

    async def check_and_rekey(self, bytes_processed: int = 0) -> bool:
        """
        Check if rekey is needed and perform it if necessary.

        Args:
            bytes_processed: Number of bytes processed in this operation

        Returns:
            bool: True if rekey was performed

        Raises:
            Exception: If rekey operation fails

        """
        if bytes_processed > 0:
            self.rekey_manager.update_bytes_processed(bytes_processed)

        if self.rekey_manager.should_rekey():
            await self.session.rekey()
            self.rekey_manager.perform_rekey()
            return True

        return False

    def get_rekey_stats(self) -> dict[str, int | float]:
        """
        Get rekey statistics.

        Returns:
            dict: Statistics about rekey operations

        """
        return self.rekey_manager.get_stats()

    def set_rekey_policy(self, policy: RekeyPolicy) -> None:
        """
        Set a new rekey policy.

        Args:
            policy: Rekey policy to use

        """
        self.rekey_manager.policy = policy



"""
Direct Connection Upgrade through Relay (DCUtR) protocol implementation.

This module implements the DCUtR protocol as specified in:
https://github.com/libp2p/specs/blob/master/relay/DCUtR.md

DCUtR enables peers behind NAT to establish direct connections
using hole punching techniques.
"""

import logging
import time
from typing import Any

from multiaddr import Multiaddr
import trio

from nanolib.abc import (
    IHost,
    INetConn,
    INetStream,
)
from nanolib.custom_types import (
    TProtocol,
)
from nanolib.peer.id import (
    ID,
)
from nanolib.peer.peerinfo import (
    PeerInfo,
)
from nanolib.relay.circuit_v2.config import (
    DEFAULT_DCUTR_READ_TIMEOUT,
    DEFAULT_DCUTR_WRITE_TIMEOUT,
    DEFAULT_DIAL_TIMEOUT,
)
from nanolib.relay.circuit_v2.nat import (
    ReachabilityChecker,
)
from nanolib.relay.circuit_v2.pb.dcutr_pb2 import (
    HolePunch,
)
from nanolib.tools.async_service import (
    Service,
)

logger = logging.getLogger(__name__)

# Protocol ID for DCUtR
PROTOCOL_ID = TProtocol("/libp2p/dcutr")

# Maximum message size for DCUtR (4KiB as per spec)
MAX_MESSAGE_SIZE = 4 * 1024

# DCUtR protocol constants
# Maximum number of hole punch attempts per peer
MAX_HOLE_PUNCH_ATTEMPTS = 5

# Delay between retry attempts
HOLE_PUNCH_RETRY_DELAY = 30  # seconds

# Maximum observed addresses to exchange
MAX_OBSERVED_ADDRS = 20


class DCUtRProtocol(Service):
    """
    DCUtRProtocol implements the Direct Connection Upgrade through Relay protocol.

    This protocol allows two NATed peers to establish direct connections through
    hole punching, after they have established an initial connection through a relay.
    """

    def __init__(
        self,
        host: IHost,
        read_timeout: int = DEFAULT_DCUTR_READ_TIMEOUT,
        write_timeout: int = DEFAULT_DCUTR_WRITE_TIMEOUT,
        dial_timeout: int = DEFAULT_DIAL_TIMEOUT,
    ):
        """
        Initialize the DCUtR protocol.

        Parameters
        ----------
        host : IHost
            The libp2p host this protocol is running on
        read_timeout : int
            Timeout for stream read operations, in seconds
        write_timeout : int
            Timeout for stream write operations, in seconds
        dial_timeout : int
            Timeout for dial operations, in seconds

        """
        super().__init__()
        self.host = host
        self.read_timeout = read_timeout
        self.write_timeout = write_timeout
        self.dial_timeout = dial_timeout
        self.event_started = trio.Event()
        self._hole_punch_attempts: dict[ID, int] = {}
        self._direct_connections: set[ID] = set()
        self._in_progress: set[ID] = set()
        self._reachability_checker = ReachabilityChecker(host)
        self._nursery: trio.Nursery | None = None

    async def run(self, *, task_status: Any = trio.TASK_STATUS_IGNORED) -> None:
        """Run the protocol service."""
        try:
            # Register the DCUtR protocol handler
            logger.debug("Registering DCUtR protocol handler")
            self.host.set_stream_handler(PROTOCOL_ID, self._handle_dcutr_stream)

            # Signal that we're ready
            self.event_started.set()

            # Start the service
            async with trio.open_nursery() as nursery:
                self._nursery = nursery
                task_status.started()
                logger.debug("DCUtR protocol service started")

                # Wait for service to be stopped
                await self.manager.wait_finished()
        finally:
            # Clean up
            try:
                # Use empty async lambda instead of None for stream handler
                async def empty_handler(_: INetStream) -> None:
                    pass

                self.host.set_stream_handler(PROTOCOL_ID, empty_handler)
                logger.debug("DCUtR protocol handler unregistered")
            except Exception as e:
                logger.error("Error unregistering DCUtR protocol handler: %s", str(e))

            # Clear state
            self._hole_punch_attempts.clear()
            self._direct_connections.clear()
            self._in_progress.clear()
            self._nursery = None

    async def _handle_dcutr_stream(self, stream: INetStream) -> None:
        """
        Handle incoming DCUtR streams.

        Parameters
        ----------
        stream : INetStream
            The incoming stream

        """
        try:
            # Get the remote peer ID
            remote_peer_id = stream.muxed_conn.peer_id
            logger.debug("Received DCUtR stream from peer %s", remote_peer_id)

            # Check if we already have a direct connection
            if await self._have_direct_connection(remote_peer_id):
                logger.debug(
                    "Already have direct connection to %s, closing stream",
                    remote_peer_id,
                )
                await stream.close()
                return

            # Check if there's already an active hole punch attempt
            if remote_peer_id in self._in_progress:
                logger.debug("Hole punch already in progress with %s", remote_peer_id)
                # Let the existing attempt continue
                await stream.close()
                return

            # Mark as in progress
            self._in_progress.add(remote_peer_id)

            try:
                # Read the CONNECT message
                with trio.fail_after(self.read_timeout):
                    msg_bytes = await stream.read(MAX_MESSAGE_SIZE)

                # Parse the message
                connect_msg = HolePunch()
                connect_msg.ParseFromString(msg_bytes)

                # Verify it's a CONNECT message
                if connect_msg.type != HolePunch.CONNECT:
                    logger.warning("Expected CONNECT message, got %s", connect_msg.type)
                    await stream.close()
                    return

                logger.debug(
                    "Received CONNECT message from %s with %d addresses",
                    remote_peer_id,
                    len(connect_msg.ObsAddrs),
                )

                # Process observed addresses from the peer
                peer_addrs = self._decode_observed_addrs(list(connect_msg.ObsAddrs))
                logger.debug("Decoded %d valid addresses from peer", len(peer_addrs))

                # Store the addresses in the peerstore
                if peer_addrs:
                    self.host.get_peerstore().add_addrs(
                        remote_peer_id, peer_addrs, 10 * 60
                    )  # 10 minute TTL

                # Send our CONNECT message with our observed addresses
                our_addrs = await self._get_observed_addrs()
                response = HolePunch()
                response.type = HolePunch.CONNECT
                response.ObsAddrs.extend(our_addrs)

                with trio.fail_after(self.write_timeout):
                    await stream.write(response.SerializeToString())

                logger.debug(
                    "Sent CONNECT response to %s with %d addresses",
                    remote_peer_id,
                    len(our_addrs),
                )

                # Wait for SYNC message
                with trio.fail_after(self.read_timeout):
                    sync_bytes = await stream.read(MAX_MESSAGE_SIZE)

                # Parse the SYNC message
                sync_msg = HolePunch()
                sync_msg.ParseFromString(sync_bytes)

                # Verify it's a SYNC message
                if sync_msg.type != HolePunch.SYNC:
                    logger.warning("Expected SYNC message, got %s", sync_msg.type)
                    await stream.close()
                    return

                logger.debug("Received SYNC message from %s", remote_peer_id)

                # Perform hole punch
                success = await self._perform_hole_punch(remote_peer_id, peer_addrs)

                if success:
                    logger.info(
                        "Successfully established direct connection with %s",
                        remote_peer_id,
                    )
                else:
                    logger.warning(
                        "Failed to establish direct connection with %s", remote_peer_id
                    )

            except trio.TooSlowError:
                logger.warning("Timeout in DCUtR protocol with peer %s", remote_peer_id)
            except Exception as e:
                logger.error(
                    "Error in DCUtR protocol with peer %s: %s", remote_peer_id, str(e)
                )
            finally:
                # Clean up
                self._in_progress.discard(remote_peer_id)
                await stream.close()

        except Exception as e:
            logger.error("Error handling DCUtR stream: %s", str(e))
            await stream.close()

    async def initiate_hole_punch(self, peer_id: ID) -> bool:
        """
        Initiate a hole punch with a peer.

        Parameters
        ----------
        peer_id : ID
            The peer to hole punch with

        Returns
        -------
        bool
            True if hole punch was successful, False otherwise

        """
        # Check if we already have a direct connection
        if await self._have_direct_connection(peer_id):
            logger.debug("Already have direct connection to %s", peer_id)
            return True

        # Check if there's already an active hole punch attempt
        if peer_id in self._in_progress:
            logger.debug("Hole punch already in progress with %s", peer_id)
            return False

        # Check if we've exceeded the maximum number of attempts
        attempts = self._hole_punch_attempts.get(peer_id, 0)
        if attempts >= MAX_HOLE_PUNCH_ATTEMPTS:
            logger.warning("Maximum hole punch attempts reached for peer %s", peer_id)
            return False

        # Mark as in progress and increment attempt counter
        self._in_progress.add(peer_id)
        self._hole_punch_attempts[peer_id] = attempts + 1

        try:
            # Open a DCUtR stream to the peer
            logger.debug("Opening DCUtR stream to peer %s", peer_id)
            stream = await self.host.new_stream(peer_id, [PROTOCOL_ID])
            if not stream:
                logger.warning("Failed to open DCUtR stream to peer %s", peer_id)
                return False

            try:
                # Send our CONNECT message with our observed addresses
                our_addrs = await self._get_observed_addrs()
                connect_msg = HolePunch()
                connect_msg.type = HolePunch.CONNECT
                connect_msg.ObsAddrs.extend(our_addrs)

                start_time = time.time()
                with trio.fail_after(self.write_timeout):
                    await stream.write(connect_msg.SerializeToString())

                logger.debug(
                    "Sent CONNECT message to %s with %d addresses",
                    peer_id,
                    len(our_addrs),
                )

                # Receive the peer's CONNECT message
                with trio.fail_after(self.read_timeout):
                    resp_bytes = await stream.read(MAX_MESSAGE_SIZE)

                # Calculate RTT
                rtt = time.time() - start_time

                # Parse the response
                resp = HolePunch()
                resp.ParseFromString(resp_bytes)

                # Verify it's a CONNECT message
                if resp.type != HolePunch.CONNECT:
                    logger.warning("Expected CONNECT message, got %s", resp.type)
                    return False

                logger.debug(
                    "Received CONNECT response from %s with %d addresses",
                    peer_id,
                    len(resp.ObsAddrs),
                )

                # Process observed addresses from the peer
                peer_addrs = self._decode_observed_addrs(list(resp.ObsAddrs))
                logger.debug("Decoded %d valid addresses from peer", len(peer_addrs))

                # Store the addresses in the peerstore
                if peer_addrs:
                    self.host.get_peerstore().add_addrs(
                        peer_id, peer_addrs, 10 * 60
                    )  # 10 minute TTL

                # Send SYNC message with timing information
                # We'll use a future time that's 2*RTT from now to ensure both sides
                # are ready
                punch_time = time.time() + (2 * rtt) + 1  # Add 1 second buffer

                sync_msg = HolePunch()
                sync_msg.type = HolePunch.SYNC

                with trio.fail_after(self.write_timeout):
                    await stream.write(sync_msg.SerializeToString())

                logger.debug("Sent SYNC message to %s", peer_id)

                # Perform the synchronized hole punch
                success = await self._perform_hole_punch(
                    peer_id, peer_addrs, punch_time
                )

                if success:
                    logger.info(
                        "Successfully established direct connection with %s", peer_id
                    )
                    return True
                else:
                    logger.warning(
                        "Failed to establish direct connection with %s", peer_id
                    )
                    return False

            except trio.TooSlowError:
                logger.warning("Timeout in DCUtR protocol with peer %s", peer_id)
                return False
            except Exception as e:
                logger.error(
                    "Error in DCUtR protocol with peer %s: %s", peer_id, str(e)
                )
                return False
            finally:
                await stream.close()

        except Exception as e:
            logger.error(
                "Error initiating hole punch with peer %s: %s", peer_id, str(e)
            )
            return False
        finally:
            self._in_progress.discard(peer_id)

        # This should never be reached, but add explicit return for type checking
        return False

    async def _perform_hole_punch(
        self, peer_id: ID, addrs: list[Multiaddr], punch_time: float | None = None
    ) -> bool:
        """
        Perform a hole punch attempt with a peer.

        Parameters
        ----------
        peer_id : ID
            The peer to hole punch with
        addrs : list[Multiaddr]
            List of addresses to try
        punch_time : Optional[float]
            Time to perform the punch (if None, do it immediately)

        Returns
        -------
        bool
            True if hole punch was successful

        """
        if not addrs:
            logger.warning("No addresses to try for hole punch with %s", peer_id)
            return False

        # If punch_time is specified, wait until that time
        if punch_time is not None:
            now = time.time()
            if punch_time > now:
                wait_time = punch_time - now
                logger.debug("Waiting %.2f seconds before hole punch", wait_time)
                await trio.sleep(wait_time)

        # Try to dial each address
        logger.debug(
            "Starting hole punch with peer %s using %d addresses", peer_id, len(addrs)
        )

        # Filter to only include non-relay addresses
        direct_addrs = [
            addr for addr in addrs if not str(addr).startswith("/p2p-circuit")
        ]

        if not direct_addrs:
            logger.warning("No direct addresses found for peer %s", peer_id)
            return False

        # Start dialing attempts in parallel
        async with trio.open_nursery() as nursery:
            for addr in direct_addrs[
                :5
            ]:  # Limit to 5 addresses to avoid too many connections
                nursery.start_soon(self._dial_peer, peer_id, addr)

        # Check if we established a direct connection
        return await self._have_direct_connection(peer_id)

    async def _dial_peer(self, peer_id: ID, addr: Multiaddr) -> None:
        """
        Attempt to dial a peer at a specific address.

        Parameters
        ----------
        peer_id : ID
            The peer to dial
        addr : Multiaddr
            The address to dial

        """
        try:
            logger.debug("Attempting to dial %s at %s", peer_id, addr)

            # Create peer info
            peer_info = PeerInfo(peer_id, [addr])

            # Try to connect with timeout
            with trio.fail_after(self.dial_timeout):
                await self.host.connect(peer_info)

            logger.info("Successfully connected to %s at %s", peer_id, addr)

            # Add to direct connections set
            self._direct_connections.add(peer_id)

        except trio.TooSlowError:
            logger.debug("Timeout dialing %s at %s", peer_id, addr)
        except Exception as e:
            logger.debug("Error dialing %s at %s: %s", peer_id, addr, str(e))

    async def _have_direct_connection(self, peer_id: ID) -> bool:
        """
        Check if we already have a direct connection to a peer.

        Parameters
        ----------
        peer_id : ID
            The peer to check

        Returns
        -------
        bool
            True if we have a direct connection, False otherwise

        """
        # Check our direct connections cache first
        if peer_id in self._direct_connections:
            return True

        # Check if the peer is connected
        network = self.host.get_network()
        conn_or_conns = network.connections.get(peer_id)
        if not conn_or_conns:
            return False

        # Handle both single connection and list of connections
        if isinstance(conn_or_conns, list):
            connections: list[INetConn] = conn_or_conns
        else:
            connections = [conn_or_conns]

        # Check if any connection is direct (not relayed)
        for conn in connections:
            # Get the transport addresses
            addrs = conn.get_transport_addresses()

            # If any address doesn't start with /p2p-circuit, it's a direct connection
            if any(not str(addr).startswith("/p2p-circuit") for addr in addrs):
                # Cache this result
                self._direct_connections.add(peer_id)
                return True

        return False

    async def _get_observed_addrs(self) -> list[bytes]:
        """
        Get our observed addresses to share with the peer.

        Returns
        -------
        List[bytes]
            List of observed addresses as bytes

        """
        # Get all listen addresses
        addrs = self.host.get_addrs()

        # Filter out relay addresses
        direct_addrs = [
            addr for addr in addrs if not str(addr).startswith("/p2p-circuit")
        ]

        # Limit the number of addresses
        if len(direct_addrs) > MAX_OBSERVED_ADDRS:
            direct_addrs = direct_addrs[:MAX_OBSERVED_ADDRS]

        # Convert to bytes
        addr_bytes = [addr.to_bytes() for addr in direct_addrs]

        return addr_bytes

    def _decode_observed_addrs(self, addr_bytes: list[bytes]) -> list[Multiaddr]:
        """
        Decode observed addresses received from a peer.

        Parameters
        ----------
        addr_bytes : List[bytes]
            The encoded addresses

        Returns
        -------
        List[Multiaddr]
            The decoded multiaddresses

        """
        result = []

        for addr_byte in addr_bytes:
            try:
                addr = Multiaddr(addr_byte)
                # Validate the address (basic check)
                if str(addr).startswith("/ip"):
                    result.append(addr)
            except Exception as e:
                logger.debug("Error decoding multiaddr: %s", str(e))

        return result



"""
Advanced TLS configuration for WebSocket transport.

This module provides comprehensive TLS configuration options including
SNI support, certificate validation, and advanced TLS features.
"""

from dataclasses import dataclass, field
from enum import Enum
import ssl

from nanolib.peer.id import ID


class TLSVersion(Enum):
    """TLS version enumeration."""

    TLS_1_0 = "TLSv1"
    TLS_1_1 = "TLSv1.1"
    TLS_1_2 = "TLSv1.2"
    TLS_1_3 = "TLSv1.3"


class CertificateValidationMode(Enum):
    """Certificate validation mode."""

    NONE = "none"  # No validation
    BASIC = "basic"  # Basic certificate validation
    STRICT = "strict"  # Strict certificate validation
    VERIFY_PEER = "verify_peer"  # Verify peer certificates


@dataclass
class SNIConfig:
    """Server Name Indication (SNI) configuration."""

    enabled: bool = True
    default_domain: str = "localhost"
    domain_mapping: dict[str, str] = field(default_factory=dict)
    wildcard_support: bool = True

    def get_domain_for_sni(self, sni_name: str | None) -> str:
        """
        Get domain for SNI name.

        Args:
            sni_name: SNI name from client

        Returns:
            Domain to use for certificate lookup

        """
        if not sni_name:
            return self.default_domain

        # Check direct mapping
        if sni_name in self.domain_mapping:
            return self.domain_mapping[sni_name]

        # Check wildcard mapping
        if self.wildcard_support:
            for pattern, domain in self.domain_mapping.items():
                if pattern.startswith("*.") and sni_name.endswith(pattern[1:]):
                    return domain

        return sni_name


@dataclass
class CertificateConfig:
    """Certificate configuration."""

    # Certificate sources
    cert_file: str | None = None
    key_file: str | None = None
    cert_data: str | None = None
    key_data: str | None = None

    # Certificate validation
    validation_mode: CertificateValidationMode = CertificateValidationMode.BASIC
    verify_peer: bool = True
    verify_hostname: bool = True

    # Certificate chain
    ca_file: str | None = None
    ca_data: str | None = None
    ca_path: str | None = None

    # Client certificates
    client_cert_file: str | None = None
    client_key_file: str | None = None
    client_cert_data: str | None = None
    client_key_data: str | None = None

    def validate(self) -> None:
        """Validate certificate configuration."""
        # Check that we have either file or data sources
        if not any([self.cert_file, self.cert_data]):
            raise ValueError("Either cert_file or cert_data must be provided")

        if not any([self.key_file, self.key_data]):
            raise ValueError("Either key_file or key_data must be provided")

        # Validate file paths if provided
        if self.cert_file and not Path(self.cert_file).exists():
            raise ValueError(f"Certificate file not found: {self.cert_file}")

        if self.key_file and not Path(self.key_file).exists():
            raise ValueError(f"Key file not found: {self.key_file}")

        if self.ca_file and not Path(self.ca_file).exists():
            raise ValueError(f"CA file not found: {self.ca_file}")


@dataclass
class TLSConfig:
    """Comprehensive TLS configuration."""

    # Basic TLS settings
    enabled: bool = True
    min_version: TLSVersion = TLSVersion.TLS_1_2
    max_version: TLSVersion = TLSVersion.TLS_1_3

    # Certificate configuration
    certificate: CertificateConfig | None = None

    # SNI configuration
    sni: SNIConfig | None = None

    # Cipher suites
    cipher_suites: list[str] | None = None
    prefer_server_ciphers: bool = True

    # Session management
    session_cache_size: int = 128
    session_timeout: int = 300  # seconds

    # Security settings
    insecure_skip_verify: bool = False
    allow_insecure_ciphers: bool = False

    # ALPN (Application-Layer Protocol Negotiation)
    alpn_protocols: list[str] = field(default_factory=lambda: ["h2", "http/1.1"])

    # Client settings
    client_auth: bool = False
    client_ca_file: str | None = None

    # Performance settings
    renegotiation: bool = False
    compression: bool = False

    def validate(self) -> None:
        """Validate TLS configuration."""
        if self.min_version.value > self.max_version.value:
            raise ValueError("min_version cannot be greater than max_version")

        if self.certificate:
            self.certificate.validate()

        if self.session_cache_size < 0:
            raise ValueError("session_cache_size must be non-negative")

        if self.session_timeout < 0:
            raise ValueError("session_timeout must be non-negative")

    def to_ssl_context(
        self, purpose: ssl.Purpose = ssl.Purpose.SERVER_AUTH
    ) -> ssl.SSLContext:
        """
        Convert TLS configuration to SSL context.

        Args:
            purpose: SSL context purpose

        Returns:
            Configured SSL context

        """
        context = ssl.create_default_context(purpose)

        # Set TLS versions
        context.minimum_version = getattr(ssl, f"TLSVersion.{self.min_version.name}")
        context.maximum_version = getattr(ssl, f"TLSVersion.{self.max_version.name}")

        # Configure certificate
        if self.certificate:
            if self.certificate.cert_file and self.certificate.key_file:
                context.load_cert_chain(
                    certfile=self.certificate.cert_file,
                    keyfile=self.certificate.key_file,
                )
            elif self.certificate.cert_data and self.certificate.key_data:
                # Create temporary files for certificate and key
                import tempfile

                with tempfile.NamedTemporaryFile(
                    mode="w", suffix=".pem", delete=False
                ) as cert_file:
                    cert_file.write(self.certificate.cert_data)
                    cert_path = cert_file.name

                with tempfile.NamedTemporaryFile(
                    mode="w", suffix=".pem", delete=False
                ) as key_file:
                    key_file.write(self.certificate.key_data)
                    key_path = key_file.name

                try:
                    context.load_cert_chain(certfile=cert_path, keyfile=key_path)
                finally:
                    # Clean up temporary files
                    import os

                    try:
                        os.unlink(cert_path)
                        os.unlink(key_path)
                    except OSError:
                        pass

            # Configure CA
            if self.certificate.ca_file:
                context.load_verify_locations(cafile=self.certificate.ca_file)
            elif self.certificate.ca_data:
                context.load_verify_locations(cadata=self.certificate.ca_data)
            elif self.certificate.ca_path:
                context.load_verify_locations(capath=self.certificate.ca_path)

        # Configure validation
        if (
            self.certificate
            and self.certificate.validation_mode == CertificateValidationMode.NONE
        ):
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
        elif (
            self.certificate
            and self.certificate.validation_mode == CertificateValidationMode.STRICT
        ):
            context.check_hostname = True
            context.verify_mode = ssl.CERT_REQUIRED
        else:
            context.check_hostname = (
                self.certificate.verify_hostname if self.certificate else True
            )
            context.verify_mode = (
                ssl.CERT_REQUIRED
                if (self.certificate and self.certificate.verify_peer)
                else ssl.CERT_NONE
            )

        # Configure cipher suites
        if self.cipher_suites:
            context.set_ciphers(":".join(self.cipher_suites))

        if self.prefer_server_ciphers:
            context.options |= ssl.OP_CIPHER_SERVER_PREFERENCE

        # Configure session management (if supported)
        if hasattr(context, "session_cache_size"):
            context.session_cache_size = self.session_cache_size  # type: ignore
        if hasattr(context, "session_timeout"):
            context.session_timeout = self.session_timeout  # type: ignore

        # Configure security options
        if self.insecure_skip_verify:
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

        if not self.allow_insecure_ciphers:
            context.options |= ssl.OP_NO_SSLv2
            context.options |= ssl.OP_NO_SSLv3
            context.options |= ssl.OP_NO_TLSv1
            context.options |= ssl.OP_NO_TLSv1_1

        # Configure ALPN
        if self.alpn_protocols:
            context.set_alpn_protocols(self.alpn_protocols)

        # Configure client authentication
        if self.client_auth:
            context.verify_mode = ssl.CERT_REQUIRED
            if self.client_ca_file:
                context.load_verify_locations(cafile=self.client_ca_file)

        # Configure performance options
        if not self.renegotiation:
            context.options |= ssl.OP_NO_RENEGOTIATION

        if not self.compression:
            context.options |= ssl.OP_NO_COMPRESSION

        return context


@dataclass
class WebSocketTLSConfig:
    """WebSocket-specific TLS configuration."""

    # Basic TLS settings
    tls_config: TLSConfig | None = None

    # AutoTLS settings
    autotls_enabled: bool = False
    autotls_domain: str = "nanolib.local"
    autotls_storage_path: str = "autotls-certs"

    # WebSocket-specific settings
    websocket_subprotocols: list[str] = field(default_factory=lambda: ["libp2p"])
    websocket_compression: bool = True
    websocket_max_message_size: int = 32 * 1024 * 1024  # 32MB

    # Connection settings
    handshake_timeout: float = 15.0
    close_timeout: float = 5.0

    def validate(self) -> None:
        """Validate WebSocket TLS configuration."""
        if self.tls_config:
            self.tls_config.validate()

        if self.handshake_timeout <= 0:
            raise ValueError("handshake_timeout must be positive")

        if self.close_timeout <= 0:
            raise ValueError("close_timeout must be positive")

        if self.websocket_max_message_size <= 0:
            raise ValueError("websocket_max_message_size must be positive")

    def get_ssl_context(
        self,
        peer_id: ID | None = None,
        sni_name: str | None = None,
    ) -> ssl.SSLContext | None:
        """
        Get SSL context for WebSocket connection.

        Args:
            peer_id: Peer ID for AutoTLS
            sni_name: SNI name for certificate selection

        Returns:
            SSL context or None if TLS is disabled

        """
        if not self.tls_config and not self.autotls_enabled:
            return None

        # Use AutoTLS if enabled
        if self.autotls_enabled and peer_id:
            from .autotls import get_autotls_manager

            manager = get_autotls_manager()
            if manager:
                domain = sni_name or self.autotls_domain
                return manager.get_ssl_context(peer_id, domain)

        # Use manual TLS configuration
        if self.tls_config:
            return self.tls_config.to_ssl_context()

        return None


# Import Path here to avoid circular imports
from pathlib import Path  # noqa: E402



"""
QUIC Connection implementation.
Manages bidirectional QUIC connections with integrated stream multiplexing.
"""

from collections import defaultdict
from collections.abc import Awaitable, Callable
import logging
import socket
import time
from typing import TYPE_CHECKING, Any, Optional

from aioquic.quic import events
from aioquic.quic.connection import QuicConnection
from aioquic.quic.events import QuicEvent
from cryptography import x509
import multiaddr
import trio

from nanolib.abc import IMuxedConn, IRawConnection
from nanolib.custom_types import TQUICStreamHandlerFn
from nanolib.peer.id import ID
from nanolib.rcmgr import Direction
from nanolib.stream_muxer.exceptions import MuxedConnUnavailable

from .exceptions import (
    QUICConnectionClosedError,
    QUICConnectionError,
    QUICConnectionTimeoutError,
    QUICErrorContext,
    QUICPeerVerificationError,
    QUICStreamError,
    QUICStreamLimitError,
    QUICStreamTimeoutError,
)
from .stream import QUICStream, StreamDirection

if TYPE_CHECKING:
    from .listener import QUICListener
    from .security import QUICTLSConfigManager
    from .transport import QUICTransport

logger = logging.getLogger(__name__)


class QUICConnection(IRawConnection, IMuxedConn):
    """
    QUIC connection implementing both raw connection and muxed connection interfaces.

    Uses aioquic's sans-IO core with trio for native async support.
    QUIC natively provides stream multiplexing, so this connection acts as both
    a raw connection (for transport layer) and muxed connection (for upper layers).

    Features:
    - Native QUIC stream multiplexing
    - Integrated libp2p TLS security with peer identity verification
    - Resource-aware stream management
    - Comprehensive error handling
    - Flow control integration
    - Connection migration support
    - Performance monitoring
    - COMPLETE connection ID management (fixes the original issue)
    """

    def __init__(
        self,
        quic_connection: QuicConnection,
        remote_addr: tuple[str, int],
        remote_peer_id: ID | None,
        local_peer_id: ID,
        is_initiator: bool,
        maddr: multiaddr.Multiaddr,
        transport: "QUICTransport",
        security_manager: Optional["QUICTLSConfigManager"] = None,
        resource_scope: Any | None = None,
        listener_socket: trio.socket.SocketType | None = None,
        listener: "QUICListener | None" = None,
        listener_connection_id: bytes | None = None,
    ):
        """
        Initialize QUIC connection with security integration.

        Args:
            quic_connection: aioquic QuicConnection instance
            remote_addr: Remote peer address
            remote_peer_id: Remote peer ID (may be None initially)
            local_peer_id: Local peer ID
            is_initiator: Whether this is the connection initiator
            maddr: Multiaddr for this connection
            transport: Parent QUIC transport
            security_manager: Security manager for TLS/certificate handling
            resource_scope: Resource manager scope for tracking
            listener_socket: Socket of listener to transmit data
            listener: Owning listener for server-side connections (enables O(1)
                Connection ID registration).
            listener_connection_id: Listener's primary Connection ID for this
                connection (used as the base CID when registering new CIDs).

        """
        self._quic = quic_connection
        self._remote_addr = remote_addr
        self._remote_peer_id = remote_peer_id
        self._local_peer_id = local_peer_id
        self.peer_id = remote_peer_id or local_peer_id
        # Derive initiator/client role from aioquic configuration when available.
        # Only trust `is_client` when it's a real bool
        #  (mocks may return truthy objects).
        is_client = getattr(
            getattr(quic_connection, "configuration", None), "is_client", None
        )
        self._is_initiator = is_client if isinstance(is_client, bool) else is_initiator
        self._maddr = maddr
        self._transport = transport
        self._security_manager = security_manager
        self._resource_scope = resource_scope
        # Owning listener context (server-side). Used to register newly-issued
        # Connection IDs in O(1) for correct packet routing.
        self._listener: "QUICListener | None" = listener
        self._listener_connection_id: bytes | None = listener_connection_id

        # Trio networking - socket may be provided by listener
        self._socket = listener_socket if listener_socket else None
        self._owns_socket = listener_socket is None
        self._connected_event = trio.Event()
        self._closed_event = trio.Event()

        self._streams: dict[int, QUICStream] = {}
        self._stream_cache: dict[int, QUICStream] = {}  # Cache for frequent lookups
        self._next_stream_id: int = self._calculate_initial_stream_id()
        self._stream_handler: TQUICStreamHandlerFn | None = None

        # Single lock for all stream operations
        self._stream_lock = trio.Lock()

        # Stream counting and limits
        self._outbound_stream_count: int = 0
        self._inbound_stream_count: int = 0

        # Stream acceptance for incoming streams
        self._stream_accept_queue: list[QUICStream] = []
        self._stream_accept_event = trio.Event()

        # Negotiation semaphores to limit concurrent multiselect negotiations
        # Separate semaphores for client (outbound) and server (inbound) to prevent
        # deadlocks where client holds all slots and server can't respond.
        # This prevents overwhelming the connection with too many simultaneous
        # negotiations, which can cause timeouts under high concurrency.
        # Limit is configurable via transport config to allow tuning for
        # different use cases. The separate client/server semaphores prevent
        # deadlocks while maintaining reasonable resource usage. In CI/CD
        # environments with limited resources, this helps prevent contention.
        # Get negotiation limit from config, defaulting to 5 if not available
        # or if it's a Mock object (in tests)
        negotiation_limit = getattr(
            self._transport._config, "NEGOTIATION_SEMAPHORE_LIMIT", 5
        )
        # Ensure it's an int (handles Mock objects in tests)
        if not isinstance(negotiation_limit, int):
            negotiation_limit = 5
        self._client_negotiation_semaphore = trio.Semaphore(negotiation_limit)
        self._server_negotiation_semaphore = trio.Semaphore(negotiation_limit)
        # Keep _negotiation_semaphore for backward compatibility (maps to client)
        self._negotiation_semaphore = self._client_negotiation_semaphore

        # Connection state
        self._closed: bool = False
        self._established: bool = False
        self._started: bool = False
        self._handshake_completed: bool = False
        self._peer_verified: bool = False

        # Security state
        self._peer_certificate: x509.Certificate | None = None
        self._handshake_events: list[events.HandshakeCompleted] = []

        # Background task management
        self._background_tasks_started: bool = False
        self._nursery: trio.Nursery | None = None
        self._event_processing_task: Any | None = None
        self.on_close: Callable[[], Awaitable[None]] | None = None
        self.event_started = trio.Event()

        self._available_connection_ids: set[bytes] = set()
        self._current_connection_id: bytes | None = None
        self._retired_connection_ids: set[bytes] = set()
        self._connection_id_sequence_numbers: set[int] = set()
        # Sequence number counter for tracking Connection IDs (inspired by quinn)
        # Starts at 0 for initial CID, increments for each new CID issued
        self._connection_id_sequence_counter: int = 0

        # Event processing control with batching
        self._event_processing_active: bool = False
        self._event_batch: list[events.QuicEvent] = []
        self._event_batch_size: int = 10
        self._last_event_time: float = 0.0

        # Set quic connection configuration
        self.CONNECTION_CLOSE_TIMEOUT = self._transport._config.CONNECTION_CLOSE_TIMEOUT
        self.MAX_INCOMING_STREAMS = self._transport._config.MAX_INCOMING_STREAMS
        self.MAX_OUTGOING_STREAMS = self._transport._config.MAX_OUTGOING_STREAMS
        self.CONNECTION_HANDSHAKE_TIMEOUT = (
            self._transport._config.CONNECTION_HANDSHAKE_TIMEOUT
        )
        self.MAX_CONCURRENT_STREAMS = self._transport._config.MAX_CONCURRENT_STREAMS
        self.STREAM_OPEN_TIMEOUT = self._transport._config.STREAM_OPEN_TIMEOUT

        # Performance monitoring thresholds
        self.SLOW_NOTIFICATION_THRESHOLD_SECONDS = 0.01  # 10ms

        # Performance and monitoring
        self._connection_start_time = time.time()
        self._stats = {
            "streams_opened": 0,
            "streams_accepted": 0,
            "streams_closed": 0,
            "streams_reset": 0,
            "bytes_sent": 0,
            "bytes_received": 0,
            "packets_sent": 0,
            "packets_received": 0,
            "connection_ids_issued": 0,
            "connection_ids_retired": 0,
            "connection_id_changes": 0,
        }

        logger.debug(
            f"Created QUIC connection to {self._remote_peer_id} "
            f"(initiator: {self._is_initiator}, addr: {self._remote_addr}, "
            f"security: {self._security_manager is not None})"
        )

    def set_listener_context(
        self, listener: "QUICListener | None", listener_connection_id: bytes | None
    ) -> None:
        """
        Set owning listener context for O(1) Connection ID registration.

        For server-side connections, the listener is responsible for packet routing.
        """
        self._listener = listener
        self._listener_connection_id = listener_connection_id

    # Resource manager integration
    def set_resource_scope(self, scope: Any) -> None:
        """
        Attach a resource scope to the connection for stream construction and cleanup.
        """
        self._resource_scope = scope

    def _calculate_initial_stream_id(self) -> int:
        """
        Calculate the initial stream ID based on QUIC specification.

        QUIC stream IDs:
        - Client-initiated bidirectional: 0, 4, 8, 12, ...
        - Server-initiated bidirectional: 1, 5, 9, 13, ...
        - Client-initiated unidirectional: 2, 6, 10, 14, ...
        - Server-initiated unidirectional: 3, 7, 11, 15, ...

        For libp2p, we primarily use bidirectional streams.
        """
        if self._is_initiator:
            return 0
        else:
            return 1

    @property
    def is_initiator(self) -> bool:  # type: ignore
        """Check if this connection is the initiator."""
        return self._is_initiator

    @property
    def is_closed(self) -> bool:
        """Check if connection is closed."""
        return self._closed

    @property
    def is_established(self) -> bool:
        """Check if connection is established (handshake completed)."""
        return self._established and self._handshake_completed

    @property
    def is_started(self) -> bool:
        """Check if connection has been started."""
        return self._started

    @property
    def is_peer_verified(self) -> bool:
        """Check if peer identity has been verified."""
        return self._peer_verified

    def multiaddr(self) -> multiaddr.Multiaddr:
        """Get the multiaddr for this connection."""
        return self._maddr

    def local_peer_id(self) -> ID:
        """Get the local peer ID."""
        return self._local_peer_id

    def remote_peer_id(self) -> ID | None:
        """Get the remote peer ID."""
        return self._remote_peer_id

    def get_connection_id_stats(self) -> dict[str, Any]:
        """Get connection ID statistics and current state."""
        return {
            "available_connection_ids": len(self._available_connection_ids),
            "current_connection_id": self._current_connection_id.hex()
            if self._current_connection_id
            else None,
            "retired_connection_ids": len(self._retired_connection_ids),
            "connection_ids_issued": self._stats["connection_ids_issued"],
            "connection_ids_retired": self._stats["connection_ids_retired"],
            "connection_id_changes": self._stats["connection_id_changes"],
            "available_cid_list": [cid.hex() for cid in self._available_connection_ids],
        }

    def get_current_connection_id(self) -> bytes | None:
        """Get the current connection ID."""
        return self._current_connection_id

    # Fast stream lookup with caching
    def _get_stream_fast(self, stream_id: int) -> QUICStream | None:
        """Get stream with caching for performance."""
        # Try cache first
        stream = self._stream_cache.get(stream_id)
        if stream is not None:
            return stream

        # Fallback to main dict
        stream = self._streams.get(stream_id)
        if stream is not None:
            self._stream_cache[stream_id] = stream

        return stream

    # Connection lifecycle methods

    async def start(self) -> None:
        """
        Start the connection and its background tasks.

        This method implements the IMuxedConn.start() interface.
        It should be called to begin processing connection events.
        """
        if self._started:
            logger.warning("Connection already started")
            return

        if self._closed:
            raise QUICConnectionError("Cannot start a closed connection")

        self._started = True
        logger.debug(f"Starting QUIC connection to {self._remote_peer_id}")

        try:
            # If this is a client connection, we need to establish the connection
            if self._is_initiator:
                await self._initiate_connection()
                # event_started will be set in connect() after connection is established
            else:
                # For server connections, we're already connected via the listener
                self._established = True
                self._connected_event.set()
                # Set event_started after connection is established for server
                self.event_started.set()

            logger.debug(f"QUIC connection to {self._remote_peer_id} started")

        except Exception as e:
            logger.error(f"Failed to start connection: {e}")
            raise QUICConnectionError(f"Connection start failed: {e}") from e

    async def _initiate_connection(self) -> None:
        """Initiate client-side connection, reusing listener socket if available."""
        try:
            with QUICErrorContext("connection_initiation", "connection"):
                if not self._socket:
                    logger.debug("Creating new socket for outbound connection")
                    self._socket = trio.socket.socket(
                        family=socket.AF_INET, type=socket.SOCK_DGRAM
                    )

                await self._socket.bind(("0.0.0.0", 0))

                self._quic.connect(self._remote_addr, now=time.time())

                # Send initial packet(s)
                await self._transmit()

                logger.debug(f"Initiated QUIC connection to {self._remote_addr}")

        except Exception as e:
            logger.error(f"Failed to initiate connection: {e}")
            raise QUICConnectionError(f"Connection initiation failed: {e}") from e

    async def connect(self, nursery: trio.Nursery) -> None:
        """
        Establish the QUIC connection using trio nursery for background tasks.

        Args:
            nursery: Trio nursery for managing connection background tasks

        """
        if self._closed:
            raise QUICConnectionClosedError("Connection is closed")

        self._nursery = nursery

        try:
            with QUICErrorContext("connection_establishment", "connection"):
                # Start the connection if not already started
                if not self._started:
                    await self.start()

                # Start background event processing
                if not self._background_tasks_started:
                    await self._start_background_tasks()

                # Wait for handshake completion with timeout
                with trio.move_on_after(
                    self.CONNECTION_HANDSHAKE_TIMEOUT
                ) as cancel_scope:
                    await self._connected_event.wait()

                if cancel_scope.cancelled_caught:
                    raise QUICConnectionTimeoutError(
                        "Connection handshake timed out after"
                        f"{self.CONNECTION_HANDSHAKE_TIMEOUT}s"
                    )

                logger.debug(
                    "QUICConnection: Verifying peer identity with security manager"
                )
                # Verify peer identity using security manager
                peer_id = await self._verify_peer_identity_with_security()

                if peer_id:
                    self.peer_id = peer_id

                logger.debug(f"QUICConnection {id(self)}: Peer identity verified")
                self._established = True
                logger.debug(f"QUIC connection established with {self._remote_peer_id}")

                # Set event_started after connection is fully established for initiator
                if self._is_initiator:
                    self.event_started.set()

        except Exception as e:
            logger.error(f"Failed to establish connection: {e}")
            await self.close()
            raise

    async def _start_background_tasks(self) -> None:
        """Start background tasks for connection management."""
        if self._background_tasks_started or not self._nursery:
            return

        self._background_tasks_started = True

        if self._is_initiator:
            self._nursery.start_soon(async_fn=self._client_packet_receiver)

        self._nursery.start_soon(async_fn=self._event_processing_loop)
        self._nursery.start_soon(async_fn=self._periodic_maintenance)

        logger.debug("Started background tasks for QUIC connection")

    async def _event_processing_loop(self) -> None:
        """Main event processing loop for the connection."""
        logger.debug(
            f"Started QUIC event processing loop for connection id: {id(self)} "
            f"and local peer id {str(self.local_peer_id())}"
        )

        try:
            consecutive_idle_iterations = 0
            while not self._closed:
                # Batch process events - returns True if events were processed
                events_processed = await self._process_quic_events_batched()

                # Handle timer events
                await self._handle_timer_events()

                # Transmit any pending data
                await self._transmit()

                # Adaptive sleep based on activity
                # When processing events: use minimal sleep (just yield) for low latency
                # When idle: use longer sleep to reduce CPU usage
                if not events_processed:
                    consecutive_idle_iterations += 1
                    # Use longer sleep when idle to reduce CPU usage
                    # Start with 1ms, increase to 10ms after several idle iterations
                    sleep_time = 0.01 if consecutive_idle_iterations > 5 else 0.001
                    await trio.sleep(sleep_time)
                else:
                    consecutive_idle_iterations = 0
                    # Minimal sleep when processing events - just yield to allow
                    # other tasks to run, but keep latency low
                    await trio.sleep(0)  # Yield without sleeping

        except Exception as e:
            logger.error(f"Error in event processing loop: {e}")
            await self._handle_connection_error(e)
        finally:
            logger.debug("QUIC event processing loop finished")

    async def _periodic_maintenance(self) -> None:
        """Perform periodic connection maintenance."""
        try:
            while not self._closed:
                # Update connection statistics
                self._update_stats()

                # Check for idle streams that can be cleaned up
                await self._cleanup_idle_streams()

                if logger.isEnabledFor(logging.DEBUG):
                    cid_stats = self.get_connection_id_stats()
                    logger.debug(f"Connection ID stats: {cid_stats}")

                # Clean cache periodically
                await self._cleanup_cache()

                # Sleep for maintenance interval
                await trio.sleep(30.0)  # 30 seconds

        except Exception as e:
            logger.error(f"Error in periodic maintenance: {e}")

    async def _cleanup_cache(self) -> None:
        """Clean up stream cache periodically to prevent memory leaks."""
        if len(self._stream_cache) > 100:  # Arbitrary threshold
            # Remove closed streams from cache
            closed_stream_ids = [
                sid for sid, stream in self._stream_cache.items() if stream.is_closed()
            ]
            for sid in closed_stream_ids:
                self._stream_cache.pop(sid, None)

    async def _client_packet_receiver(self) -> None:
        """Receive packets for client connections."""
        logger.debug("Starting client packet receiver")
        logger.debug("Started QUIC client packet receiver")

        try:
            while not self._closed and self._socket:
                try:
                    # Receive UDP packets
                    data, addr = await self._socket.recvfrom(65536)
                    logger.debug(f"Client received {len(data)} bytes from {addr}")

                    # Feed packet to QUIC connection
                    self._quic.receive_datagram(data, addr, now=time.time())

                    # Batch process events
                    await self._process_quic_events_batched()

                    # Send any response packets
                    await self._transmit()

                except trio.ClosedResourceError:
                    logger.debug("Client socket closed")
                    break
                except Exception as e:
                    logger.error(f"Error receiving client packet: {e}")
                    await trio.sleep(0.01)

        except trio.Cancelled:
            logger.debug("Client packet receiver cancelled")
            raise
        finally:
            logger.debug("Client packet receiver terminated")

    # Security and identity methods

    async def _verify_peer_identity_with_security(self) -> ID | None:
        """
        Verify peer identity using integrated security manager.

        Raises:
            QUICPeerVerificationError: If peer verification fails

        """
        if not self._security_manager:
            logger.debug("No security manager available for peer verification")
            return None

        try:
            # Extract peer certificate from TLS handshake
            await self._extract_peer_certificate()

            if not self._peer_certificate:
                logger.debug("No peer certificate available for verification")
                return None

            # Validate certificate format and accessibility
            if not self._validate_peer_certificate():
                logger.debug("Validation Failed for peer cerificate")
                raise QUICPeerVerificationError("Peer certificate validation failed")

            # Verify peer identity using security manager
            verified_peer_id = self._security_manager.verify_peer_identity(
                self._peer_certificate,
                self._remote_peer_id,  # Expected peer ID for outbound connections
            )

            # Update peer ID if it wasn't known (inbound connections)
            if not self._remote_peer_id:
                self._remote_peer_id = verified_peer_id
                logger.debug(f"Discovered peer ID from certificate: {verified_peer_id}")
            elif self._remote_peer_id != verified_peer_id:
                raise QUICPeerVerificationError(
                    f"Peer ID mismatch: expected {self._remote_peer_id}, "
                    "got {verified_peer_id}"
                )

            self._peer_verified = True
            logger.debug(f"Peer identity verified successfully: {verified_peer_id}")

            return verified_peer_id

        except QUICPeerVerificationError:
            # Re-raise verification errors as-is
            raise
        except Exception as e:
            # Wrap other errors in verification error
            raise QUICPeerVerificationError(f"Peer verification failed: {e}") from e

    async def _extract_peer_certificate(self) -> None:
        """Extract peer certificate from completed TLS handshake."""
        try:
            # Get peer certificate from aioquic TLS context
            if self._quic.tls:
                tls_context = self._quic.tls

                if tls_context._peer_certificate:
                    # aioquic stores the peer certificate as cryptography
                    # x509.Certificate
                    self._peer_certificate = tls_context._peer_certificate
                    logger.debug(
                        f"Extracted peer certificate: {self._peer_certificate.subject}"
                    )
                else:
                    logger.debug("No peer certificate found in TLS context")

            else:
                logger.debug("No TLS context available for certificate extraction")

        except Exception as e:
            logger.warning(f"Failed to extract peer certificate: {e}")

            # Try alternative approach - check if certificate is in handshake events
            try:
                # Some versions of aioquic might expose certificate differently
                config = self._quic.configuration
                if hasattr(config, "certificate") and config.certificate:
                    # This would be the local certificate, not peer certificate
                    # but we can use it for debugging
                    logger.debug("Found local certificate in configuration")

            except Exception as inner_e:
                logger.error(
                    f"Alternative certificate extraction also failed: {inner_e}"
                )

    async def get_peer_certificate(self) -> x509.Certificate | None:
        """
        Get the peer's TLS certificate.

        Returns:
            The peer's X.509 certificate, or None if not available

        """
        # If we don't have a certificate yet, try to extract it
        if not self._peer_certificate and self._handshake_completed:
            await self._extract_peer_certificate()

        return self._peer_certificate

    def _validate_peer_certificate(self) -> bool:
        """
        Validate that the peer certificate is properly formatted and accessible.

        Returns:
            True if certificate is valid and accessible, False otherwise

        """
        if not self._peer_certificate:
            return False

        try:
            # Basic validation - try to access certificate properties
            subject = self._peer_certificate.subject
            serial_number = self._peer_certificate.serial_number

            logger.debug(
                f"Certificate validation - Subject: {subject}, Serial: {serial_number}"
            )
            return True

        except Exception as e:
            logger.error(f"Certificate validation failed: {e}")
            return False

    def get_security_manager(self) -> Optional["QUICTLSConfigManager"]:
        """Get the security manager for this connection."""
        return self._security_manager

    def get_security_info(self) -> dict[str, Any]:
        """Get security-related information about the connection."""
        info: dict[str, bool | Any | None] = {
            "peer_verified": self._peer_verified,
            "handshake_complete": self._handshake_completed,
            "peer_id": str(self._remote_peer_id) if self._remote_peer_id else None,
            "local_peer_id": str(self._local_peer_id),
            "is_initiator": self._is_initiator,
            "has_certificate": self._peer_certificate is not None,
            "security_manager_available": self._security_manager is not None,
        }

        # Add certificate details if available
        if self._peer_certificate:
            try:
                info.update(
                    {
                        "certificate_subject": str(self._peer_certificate.subject),
                        "certificate_issuer": str(self._peer_certificate.issuer),
                        "certificate_serial": str(self._peer_certificate.serial_number),
                        "certificate_not_before": (
                            self._peer_certificate.not_valid_before.isoformat()
                        ),
                        "certificate_not_after": (
                            self._peer_certificate.not_valid_after.isoformat()
                        ),
                    }
                )
            except Exception as e:
                info["certificate_error"] = str(e)

        # Add TLS context debug info
        try:
            if hasattr(self._quic, "tls") and self._quic.tls:
                tls_info = {
                    "tls_context_available": True,
                    "tls_state": getattr(self._quic.tls, "state", None),
                }

                # Check for peer certificate in TLS context
                if hasattr(self._quic.tls, "_peer_certificate"):
                    tls_info["tls_peer_certificate_available"] = (
                        self._quic.tls._peer_certificate is not None
                    )

                info["tls_debug"] = tls_info
            else:
                info["tls_debug"] = {"tls_context_available": False}

        except Exception as e:
            info["tls_debug"] = {"error": str(e)}

        return info

    # Stream management methods (IMuxedConn interface)

    async def open_stream(self, timeout: float | None = None) -> QUICStream:
        """
        Open a new outbound stream

        Args:
            timeout: Timeout for stream creation
                (defaults to STREAM_OPEN_TIMEOUT from config)

        Returns:
            New QUIC stream

        Raises:
            QUICStreamLimitError: Too many concurrent streams
            QUICConnectionClosedError: Connection is closed
            QUICStreamTimeoutError: Stream creation timed out

        """
        if self._closed:
            raise QUICConnectionClosedError("Connection is closed")

        if not self._started:
            raise QUICConnectionError("Connection not started")

        # Use config timeout if not specified
        if timeout is None:
            timeout = self.STREAM_OPEN_TIMEOUT

        # Use single lock for all stream operations
        with trio.move_on_after(timeout):
            async with self._stream_lock:
                # Check stream limits inside lock
                if self._outbound_stream_count >= self.MAX_OUTGOING_STREAMS:
                    raise QUICStreamLimitError(
                        "Maximum outbound streams "
                        f"({self.MAX_OUTGOING_STREAMS}) reached"
                    )

                # Generate next stream ID
                stream_id = self._next_stream_id
                self._next_stream_id += 4  # Increment by 4 for bidirectional streams

                # Acquire stream scope from resource manager if available
                stream_scope = None
                if self._resource_scope and hasattr(
                    self._resource_scope, "_resource_manager"
                ):
                    try:
                        stream_scope = (
                            self._resource_scope._resource_manager.open_stream(
                                peer_id=self._remote_peer_id,
                                direction=Direction.OUTBOUND,
                            )
                        )
                    except Exception as e:
                        logger.warning(f"Failed to acquire stream scope: {e}")
                        raise QUICStreamLimitError(f"Resource limit exceeded: {e}")

                stream = QUICStream(
                    connection=self,
                    stream_id=stream_id,
                    direction=StreamDirection.OUTBOUND,
                    resource_scope=stream_scope,
                    remote_addr=self._remote_addr,
                )

                self._streams[stream_id] = stream
                self._stream_cache[stream_id] = stream  # Add to cache

                self._outbound_stream_count += 1
                self._stats["streams_opened"] += 1

                logger.debug(f"Opened outbound QUIC stream {stream_id}")
                return stream

        raise QUICStreamTimeoutError(f"Stream creation timed out after {timeout}s")

    async def accept_stream(self, timeout: float | None = None) -> QUICStream:
        """
        Accept incoming stream.

        Args:
            timeout: Optional timeout. If None, waits indefinitely.

        """
        if self._closed:
            raise QUICConnectionClosedError("Connection is closed")

        if timeout is not None:
            with trio.move_on_after(timeout):
                return await self._accept_stream_impl()
            # Timeout occurred
            if self._closed_event.is_set() or self._closed:
                raise MuxedConnUnavailable("QUIC connection closed during timeout")
            else:
                raise QUICStreamTimeoutError(
                    f"Stream accept timed out after {timeout}s"
                )
        else:
            # No timeout - wait indefinitely
            return await self._accept_stream_impl()

    async def _accept_stream_impl(self) -> QUICStream:
        while True:
            if self._closed:
                raise MuxedConnUnavailable("QUIC connection is closed")

            # Use single lock for stream acceptance
            async with self._stream_lock:
                if self._stream_accept_queue:
                    stream = self._stream_accept_queue.pop(0)
                    logger.debug(f"Accepted inbound stream {stream.stream_id}")
                    return stream

                # If the event is already set but the queue is empty, reset it to avoid
                # a busy loop that can starve the QUIC event loop under load.
                if self._stream_accept_event.is_set():
                    self._stream_accept_event = trio.Event()
                    continue

                accept_event = self._stream_accept_event

            if self._closed:
                raise MuxedConnUnavailable("Connection closed while accepting stream")

            # Wait for new streams indefinitely
            await accept_event.wait()

        raise QUICConnectionError("Error occurred while waiting to accept stream")

    def set_stream_handler(self, handler_function: TQUICStreamHandlerFn) -> None:
        """
        Set handler for incoming streams.

        Args:
            handler_function: Function to handle new incoming streams

        """
        self._stream_handler = handler_function
        logger.debug("Set stream handler for incoming streams")

    def _remove_stream(self, stream_id: int) -> None:
        """
        Remove stream from connection registry.
        Called by stream cleanup process.
        """
        if stream_id in self._streams:
            stream = self._streams.pop(stream_id)
            # Remove from cache too
            self._stream_cache.pop(stream_id, None)

            # Update stream counts asynchronously
            async def update_counts() -> None:
                async with self._stream_lock:
                    if stream.direction == StreamDirection.OUTBOUND:
                        self._outbound_stream_count = max(
                            0, self._outbound_stream_count - 1
                        )
                    else:
                        self._inbound_stream_count = max(
                            0, self._inbound_stream_count - 1
                        )
                    self._stats["streams_closed"] += 1

            # Schedule count update if we're in a trio context
            if self._nursery:
                self._nursery.start_soon(update_counts)

            logger.debug(f"Removed stream {stream_id} from connection")

    # Batched event processing to reduce overhead
    async def _process_quic_events_batched(self) -> bool:
        """
        Process QUIC events in batches for better performance.

        Returns:
            True if events were processed, False if no events available

        """
        if self._event_processing_active:
            return False  # Prevent recursion

        self._event_processing_active = True
        result = False  # Default to False if no events processed

        try:
            current_time = time.time()
            events_processed = 0

            # Collect events into batch
            while events_processed < self._event_batch_size:
                event = self._quic.next_event()
                if event is None:
                    break

                self._event_batch.append(event)
                events_processed += 1

            # Process batch if we have events or timeout
            if self._event_batch and (
                len(self._event_batch) >= self._event_batch_size
                or current_time - self._last_event_time > 0.01  # 10ms timeout
            ):
                await self._process_event_batch()
                self._event_batch.clear()
                self._last_event_time = current_time
                result = True
        finally:
            self._event_processing_active = False

        return result

    async def _process_event_batch(self) -> None:
        """Process a batch of events efficiently."""
        if not self._event_batch:
            return

        # Group events by type for batch processing where possible
        events_by_type: defaultdict[str, list[QuicEvent]] = defaultdict(list)
        for event in self._event_batch:
            events_by_type[type(event).__name__].append(event)

        # Process events by type
        for event_type, event_list in events_by_type.items():
            if event_type == type(events.StreamDataReceived).__name__:
                # Filter to only StreamDataReceived events
                stream_data_events = [
                    e for e in event_list if isinstance(e, events.StreamDataReceived)
                ]
                await self._handle_stream_data_batch(stream_data_events)
            else:
                # Process other events individually
                for event in event_list:
                    await self._handle_quic_event(event)

        logger.debug(f"Processed batch of {len(self._event_batch)} events")

    async def _handle_stream_data_batch(
        self, events_list: list[events.StreamDataReceived]
    ) -> None:
        """Handle stream data events in batch for better performance."""
        # Group by stream ID
        events_by_stream: defaultdict[int, list[QuicEvent]] = defaultdict(list)
        for event in events_list:
            events_by_stream[event.stream_id].append(event)

        # Process each stream's events
        for stream_id, stream_events in events_by_stream.items():
            stream = self._get_stream_fast(stream_id)  # Use fast lookup

            if not stream:
                if self._is_incoming_stream(stream_id):
                    try:
                        stream = await self._create_inbound_stream(stream_id)
                    except QUICStreamLimitError:
                        # Reset stream if we can't handle it
                        self._quic.reset_stream(stream_id, error_code=0x04)
                        await self._transmit()
                        continue
                else:
                    # Common benign case: we closed and removed a locally-initiated
                    # stream wrapper, then received a late FIN-only event.
                    fin_only = True
                    for e in stream_events:
                        data = getattr(e, "data", b"")
                        end_stream = getattr(e, "end_stream", False)
                        if data or not end_stream:
                            fin_only = False
                            break
                    if fin_only:
                        logger.debug(
                            "Ignoring late FIN on closed outbound stream %s "
                            "(is_initiator=%s, quic.is_client=%s)",
                            stream_id,
                            self._is_initiator,
                            getattr(
                                getattr(self._quic, "configuration", None),
                                "is_client",
                                None,
                            ),
                        )
                        continue
                    is_client = getattr(
                        getattr(self._quic, "configuration", None), "is_client", None
                    )
                    parity = "even" if stream_id % 2 == 0 else "odd"
                    logger.error(
                        f"Unexpected outbound stream {stream_id} in data event "
                        f"(parity={parity}, is_initiator={self._is_initiator}, "
                        f"quic.is_client={is_client}, streams={len(self._streams)}, "
                        f"cached={stream_id in self._stream_cache})"
                    )
                    continue

            # Process all events for this stream
            for received_event in stream_events:
                if hasattr(received_event, "data"):
                    self._stats["bytes_received"] += len(received_event.data)  # type: ignore

                    if hasattr(received_event, "end_stream"):
                        await stream.handle_data_received(
                            received_event.data,  # type: ignore
                            received_event.end_stream,  # type: ignore
                        )

    async def _create_inbound_stream(self, stream_id: int) -> QUICStream:
        """Create inbound stream with proper limit checking."""
        async with self._stream_lock:
            # Double-check stream doesn't exist
            existing_stream = self._streams.get(stream_id)
            if existing_stream:
                return existing_stream

            # Check limits
            if self._inbound_stream_count >= self.MAX_INCOMING_STREAMS:
                logger.warning(f"Rejecting inbound stream {stream_id}: limit reached")
                raise QUICStreamLimitError("Too many inbound streams")

            # Acquire stream scope from resource manager if available
            stream_scope = None
            if self._resource_scope and hasattr(
                self._resource_scope, "_resource_manager"
            ):
                try:
                    stream_scope = self._resource_scope._resource_manager.open_stream(
                        peer_id=self._remote_peer_id, direction=Direction.INBOUND
                    )
                except Exception as e:
                    logger.warning(f"Failed to acquire stream scope: {e}")
                    raise QUICStreamLimitError(f"Resource limit exceeded: {e}")

            # Create stream
            stream = QUICStream(
                connection=self,
                stream_id=stream_id,
                direction=StreamDirection.INBOUND,
                resource_scope=stream_scope,
                remote_addr=self._remote_addr,
            )

            self._streams[stream_id] = stream
            self._stream_cache[stream_id] = stream  # Add to cache
            self._inbound_stream_count += 1
            self._stats["streams_accepted"] += 1

            # Add to accept queue
            self._stream_accept_queue.append(stream)
            self._stream_accept_event.set()

            logger.debug(f"Created inbound stream {stream_id}")
            return stream

    async def _process_quic_events(self) -> None:
        """Process all pending QUIC events."""
        # Delegate to batched processing for better performance
        await self._process_quic_events_batched()

    async def _handle_quic_event(self, event: events.QuicEvent) -> None:
        """
        Handle a single QUIC event with complete event type coverage.

        NOTE: This is called by the connection's event loop for established connections.
        For pending connections, the listener's _process_quic_events handles events
        until the connection is promoted. This separation prevents double-processing
        of events.
        """
        logger.debug(f"Handling QUIC event: {type(event).__name__}")

        try:
            if isinstance(event, events.ConnectionTerminated):
                await self._handle_connection_terminated(event)
            elif isinstance(event, events.HandshakeCompleted):
                await self._handle_handshake_completed(event)
            elif isinstance(event, events.StreamDataReceived):
                await self._handle_stream_data(event)
            elif isinstance(event, events.StreamReset):
                await self._handle_stream_reset(event)
            elif isinstance(event, events.DatagramFrameReceived):
                await self._handle_datagram_received(event)
            # Connection ID event handlers - critical for proper packet routing
            elif isinstance(event, events.ConnectionIdIssued):
                await self._handle_connection_id_issued(event)
            elif isinstance(event, events.ConnectionIdRetired):
                await self._handle_connection_id_retired(event)
            # Additional event handlers for completeness
            elif isinstance(event, events.PingAcknowledged):
                await self._handle_ping_acknowledged(event)
            elif isinstance(event, events.ProtocolNegotiated):
                await self._handle_protocol_negotiated(event)
            elif isinstance(event, events.StopSendingReceived):
                await self._handle_stop_sending_received(event)
            else:
                logger.debug(f"Unhandled QUIC event type: {type(event).__name__}")

        except Exception as e:
            logger.error(f"Error handling QUIC event {type(event).__name__}: {e}")

    async def _handle_connection_id_issued(
        self, event: events.ConnectionIdIssued
    ) -> None:
        """
        Handle new connection ID issued by peer.

        This is the CRITICAL missing functionality that was causing your issue!
        Tracks sequence numbers for proper Connection ID retirement ordering
        (inspired by quinn).
        """
        new_connection_id = event.connection_id

        # Increment sequence counter for this new Connection ID
        sequence = self._connection_id_sequence_counter
        self._connection_id_sequence_counter += 1
        self._connection_id_sequence_numbers.add(sequence)

        logger.debug(
            f"NEW CONNECTION ID ISSUED: {new_connection_id.hex()} (sequence {sequence})"
        )

        # Add to available connection IDs
        self._available_connection_ids.add(new_connection_id)

        # If we don't have a current connection ID, use this one
        if self._current_connection_id is None:
            self._current_connection_id = new_connection_id
            logger.debug(f"Set current connection ID to: {new_connection_id.hex()}")

        # CRITICAL FIX: Notify listener to register this new Connection ID
        # with sequence number. This ensures packets with the new Connection ID
        # can be routed correctly
        await self._notify_listener_of_new_connection_id(new_connection_id, sequence)

        # Update statistics
        self._stats["connection_ids_issued"] += 1

        logger.debug(f"Available connection IDs: {len(self._available_connection_ids)}")

    async def _notify_listener_of_new_connection_id(
        self, new_connection_id: bytes, sequence: int
    ) -> None:
        """
        Notify the parent listener to register a new Connection ID.

        This is critical for proper packet routing when the peer issues
        new Connection IDs after the handshake completes.

        Args:
            new_connection_id: New Connection ID to register
            sequence: Sequence number for this Connection ID

        """
        notification_start = time.time()
        try:
            if self._listener and self._listener_connection_id:
                await self._listener._registry.add_connection_id(
                    new_connection_id, self._listener_connection_id, sequence
                )
                notification_duration = time.time() - notification_start
                if notification_duration > self.SLOW_NOTIFICATION_THRESHOLD_SECONDS:
                    logger.debug(
                        f"Slow Connection ID notification: "
                        f"{notification_duration * 1000:.2f}ms "
                        f"for Connection ID {new_connection_id.hex()[:8]}"
                    )
                return

            if not self._transport:
                return

            # Find the listener that owns this connection
            for listener in self._transport._listeners:
                # Find this connection in the listener's registry
                cids = await listener._registry.get_all_cids_for_connection(self)
                if cids:
                    # Use the first Connection ID found as the original Connection ID
                    original_connection_id = cids[0]
                    # Register new Connection ID using the registry with sequence number
                    await listener._registry.add_connection_id(
                        new_connection_id, original_connection_id, sequence
                    )
                    notification_duration = time.time() - notification_start
                    if notification_duration > 0.01:  # Log slow notifications (>10ms)
                        logger.debug(
                            f"Slow Connection ID notification: "
                            f"{notification_duration * 1000:.2f}ms "
                            f"for Connection ID {new_connection_id.hex()[:8]}"
                        )
                    logger.debug(
                        f"Registered new Connection ID {new_connection_id.hex()[:8]} "
                        f"(sequence {sequence}) for connection "
                        f"{original_connection_id.hex()[:8]}"
                    )
                    return

            logger.debug(
                f"Could not find listener to register new Connection ID "
                f"{new_connection_id.hex()[:8]}"
            )
        except Exception as e:
            logger.error(f"Error notifying listener of new CID: {e}")

    async def _handle_connection_id_retired(
        self, event: events.ConnectionIdRetired
    ) -> None:
        """
        Handle connection ID retirement.

        This handles when the peer tells us to stop using a connection ID.
        The listener will handle proper retirement ordering via the registry.
        """
        logger.debug(f"CONNECTION ID RETIRED: {event.connection_id.hex()}")

        # Remove from available IDs and add to retired set
        self._available_connection_ids.discard(event.connection_id)
        self._retired_connection_ids.add(event.connection_id)

        # If this was our current connection ID, switch to another
        if self._current_connection_id == event.connection_id:
            if self._available_connection_ids:
                self._current_connection_id = next(iter(self._available_connection_ids))
                if self._current_connection_id:
                    logger.debug(
                        "Switching to new connection ID: "
                        f"{self._current_connection_id.hex()}"
                    )
                    self._stats["connection_id_changes"] += 1
                else:
                    logger.warning("⚠️ No available connection IDs after retirement!")
            else:
                self._current_connection_id = None
                logger.warning("⚠️ No available connection IDs after retirement!")

        # Update statistics
        self._stats["connection_ids_retired"] += 1

        # Note: The listener's _process_quic_events() will handle proper
        # retirement ordering via the registry's
        # retire_connection_ids_by_sequence_range()

    async def _handle_ping_acknowledged(self, event: events.PingAcknowledged) -> None:
        """Handle ping acknowledgment."""
        logger.debug(f"Ping acknowledged: uid={event.uid}")

    async def _handle_protocol_negotiated(
        self, event: events.ProtocolNegotiated
    ) -> None:
        """Handle protocol negotiation completion."""
        logger.debug(f"Protocol negotiated: {event.alpn_protocol}")

    async def _handle_stop_sending_received(
        self, event: events.StopSendingReceived
    ) -> None:
        """Handle stop sending request from peer."""
        logger.debug(
            "Stop sending received: "
            f"stream_id={event.stream_id}, error_code={event.error_code}"
        )

        # Use fast lookup
        stream = self._get_stream_fast(event.stream_id)
        if stream:
            # Handle stop sending on the stream if method exists
            await stream.handle_stop_sending(event.error_code)

    async def _handle_handshake_completed(
        self, event: events.HandshakeCompleted
    ) -> None:
        """Handle handshake completion with security integration."""
        logger.debug("QUIC handshake completed")
        self._handshake_completed = True

        # Store handshake event for security verification
        self._handshake_events.append(event)

        # Try to extract certificate information after handshake
        await self._extract_peer_certificate()

        logger.debug("✅ Setting connected event")
        self._connected_event.set()

    async def _handle_connection_terminated(
        self, event: events.ConnectionTerminated
    ) -> None:
        """Handle connection termination."""
        logger.debug(f"QUIC connection terminated: {event.reason_phrase}")

        # Close all streams
        for stream in list(self._streams.values()):
            if event.error_code:
                await stream.handle_reset(event.error_code)
            else:
                await stream.close()

        self._streams.clear()
        self._stream_cache.clear()  # Clear cache too
        self._closed = True
        self._closed_event.set()

        self._stream_accept_event.set()
        logger.debug(f"Woke up pending accept_stream() calls, {id(self)}")

        await self._notify_parent_of_termination()

    async def _handle_stream_data(self, event: events.StreamDataReceived) -> None:
        """Handle stream data events - create streams and add to accept queue."""
        stream_id = event.stream_id
        self._stats["bytes_received"] += len(event.data)

        try:
            # Use fast lookup
            stream = self._get_stream_fast(stream_id)

            if not stream:
                if self._is_incoming_stream(stream_id):
                    logger.debug(f"Creating new incoming stream {stream_id}")
                    stream = await self._create_inbound_stream(stream_id)
                else:
                    if not event.data and event.end_stream:
                        logger.debug(
                            "Ignoring late FIN on closed outbound stream %s "
                            "(is_initiator=%s, quic.is_client=%s)",
                            stream_id,
                            self._is_initiator,
                            getattr(
                                getattr(self._quic, "configuration", None),
                                "is_client",
                                None,
                            ),
                        )
                        return
                    is_client = getattr(
                        getattr(self._quic, "configuration", None), "is_client", None
                    )
                    parity = "even" if stream_id % 2 == 0 else "odd"
                    logger.error(
                        f"Unexpected outbound stream {stream_id} in data event "
                        f"(parity={parity}, is_initiator={self._is_initiator}, "
                        f"quic.is_client={is_client}, streams={len(self._streams)}, "
                        f"cached={stream_id in self._stream_cache})"
                    )
                    return

            await stream.handle_data_received(event.data, event.end_stream)

        except Exception as e:
            logger.error(f"Error handling stream data for stream {stream_id}: {e}")
            logger.debug(f"❌ STREAM_DATA: Error: {e}")

    async def _get_or_create_stream(self, stream_id: int) -> QUICStream:
        """Get existing stream or create new inbound stream."""
        # Use fast lookup
        stream = self._get_stream_fast(stream_id)
        if stream:
            return stream

        # Check if this is an incoming stream
        is_incoming = self._is_incoming_stream(stream_id)

        if not is_incoming:
            # This shouldn't happen - outbound streams should be created by open_stream
            raise QUICStreamError(
                f"Received data for unknown outbound stream {stream_id}"
            )

        # Create new inbound stream
        return await self._create_inbound_stream(stream_id)

    def _is_incoming_stream(self, stream_id: int) -> bool:
        """
        Determine if a stream ID represents an incoming stream.

        For bidirectional streams:
        - Even IDs are client-initiated
        - Odd IDs are server-initiated
        """
        if self._is_initiator:
            # We're the client, so odd stream IDs are incoming
            return stream_id % 2 == 1
        else:
            # We're the server, so even stream IDs are incoming
            return stream_id % 2 == 0

    async def _handle_stream_reset(self, event: events.StreamReset) -> None:
        """Stream reset handling."""
        stream_id = event.stream_id
        self._stats["streams_reset"] += 1

        # Use fast lookup
        stream = self._get_stream_fast(stream_id)
        if stream:
            try:
                await stream.handle_reset(event.error_code)
                logger.debug(
                    f"Handled reset for stream {stream_id}"
                    f"with error code {event.error_code}"
                )
            except Exception as e:
                logger.error(f"Error handling stream reset for {stream_id}: {e}")
                # Force remove the stream
                self._remove_stream(stream_id)
        else:
            logger.debug(f"Received reset for unknown stream {stream_id}")

    async def _handle_datagram_received(
        self, event: events.DatagramFrameReceived
    ) -> None:
        """Handle datagram frame (if using QUIC datagrams)."""
        logger.debug(f"Datagram frame received: size={len(event.data)}")
        # For now, just log. Could be extended for custom datagram handling

    async def _handle_timer_events(self) -> None:
        """Handle QUIC timer events."""
        timer = self._quic.get_timer()
        if timer is not None:
            now = time.time()
            if timer <= now:
                self._quic.handle_timer(now=now)

    # Network transmission

    async def _transmit(self) -> None:
        """Transmit pending QUIC packets using available socket."""
        sock = self._socket
        if not sock:
            logger.debug("No socket to transmit")
            return

        try:
            current_time = time.time()
            datagrams = self._quic.datagrams_to_send(now=current_time)

            # Batch stats updates
            packet_count = 0
            total_bytes = 0

            for data, addr in datagrams:
                await sock.sendto(data, addr)
                packet_count += 1
                total_bytes += len(data)

            # Update stats in batch
            if packet_count > 0:
                self._stats["packets_sent"] += packet_count
                self._stats["bytes_sent"] += total_bytes

        except Exception as e:
            logger.error(f"Transmission error: {e}")
            await self._handle_connection_error(e)

    # Additional methods for stream data processing
    async def _process_quic_event(self, event: events.QuicEvent) -> None:
        """Process a single QUIC event."""
        await self._handle_quic_event(event)

    async def _transmit_pending_data(self) -> None:
        """Transmit any pending data."""
        await self._transmit()

    # Error handling

    async def _handle_connection_error(self, error: Exception) -> None:
        """Handle connection-level errors."""
        logger.error(f"Connection error: {error}")

        if not self._closed:
            try:
                await self.close()
            except Exception as close_error:
                logger.error(f"Error during connection close: {close_error}")

    # Connection close

    async def close(self) -> None:
        """Connection close with proper stream cleanup."""
        if self._closed:
            return

        self._closed = True
        logger.debug(f"Closing QUIC connection to {self._remote_peer_id}")

        try:
            # Close all streams gracefully, but limit concurrency to prevent
            # excessive CPU usage when many streams fail simultaneously
            stream_close_tasks = []
            for stream in list(self._streams.values()):
                if stream.can_write() or stream.can_read():
                    stream_close_tasks.append(stream.close)

            if stream_close_tasks and self._nursery:
                try:
                    # Close streams in batches to prevent overwhelming the system
                    # when many streams fail simultaneously (e.g., 100 streams)
                    batch_size = 20  # Close streams in batches of 20
                    with trio.move_on_after(self.CONNECTION_CLOSE_TIMEOUT):
                        for i in range(0, len(stream_close_tasks), batch_size):
                            batch = stream_close_tasks[i : i + batch_size]
                            async with trio.open_nursery() as close_nursery:
                                for task in batch:
                                    close_nursery.start_soon(task)
                except Exception as e:
                    logger.warning(f"Error during graceful stream close: {e}")
                    # Force reset remaining streams quickly without batching
                    # to prevent resource leaks
                    for stream in list(self._streams.values()):
                        try:
                            # Use move_on_after to prevent hanging on stuck streams
                            with trio.move_on_after(0.1):  # 100ms per stream max
                                await stream.reset(error_code=0)
                        except Exception:
                            pass

            if self.on_close:
                await self.on_close()

            # Close QUIC connection
            self._quic.close()

            if self._socket:
                await self._transmit()  # Send close frames

            # Close socket
            if self._socket and self._owns_socket:
                self._socket.close()
                self._socket = None

            self._streams.clear()
            self._stream_cache.clear()  # Clear cache
            self._closed_event.set()

            logger.debug(f"QUIC connection to {self._remote_peer_id} closed")

            # Release resource scope if present
            try:
                scope = getattr(self, "_resource_scope", None)
                if scope is not None and hasattr(scope, "done"):
                    scope.done()
                    self._resource_scope = None
            except Exception:
                pass

        except Exception as e:
            logger.error(f"Error during connection close: {e}")

    async def _notify_parent_of_termination(self) -> None:
        """
        Notify the parent listener/transport to remove this connection from tracking.

        This ensures that terminated connections are cleaned up from the
        'established connections' list.
        """
        try:
            if self._transport:
                await self._transport._cleanup_terminated_connection(self)
                logger.debug("Notified transport of connection termination")
                # Also try to remove from listeners
                for listener in self._transport._listeners:
                    try:
                        await listener._remove_connection_by_object(self)
                        logger.debug(
                            "Found and notified listener of connection termination"
                        )
                        break
                    except Exception:
                        continue
                return

            # Method 4: Use connection ID if we have one (most reliable)
            if self._current_connection_id:
                await self._cleanup_by_connection_id(self._current_connection_id)
                return

            logger.warning(
                "Could not notify parent of connection termination - no"
                f" parent reference found for conn host {self._quic.host_cid.hex()}"
            )

        except Exception as e:
            logger.error(f"Error notifying parent of connection termination: {e}")

    async def _cleanup_by_connection_id(self, connection_id: bytes) -> None:
        """Cleanup using connection ID as a fallback method."""
        try:
            for listener in self._transport._listeners:
                # Find this connection in the listener's registry
                cids = await listener._registry.get_all_cids_for_connection(self)
                if cids:
                    # Remove using the first Connection ID found
                    tracked_cid = cids[0]
                    await listener._remove_connection(tracked_cid)
                    logger.debug(f"Removed connection {tracked_cid.hex()}")
                    return

            logger.debug("Fallback cleanup by connection ID completed")
        except Exception as e:
            logger.error(f"Error in fallback cleanup: {e}")

    # IRawConnection interface (for compatibility)

    def get_remote_address(self) -> tuple[str, int]:
        return self._remote_addr

    async def write(self, data: bytes) -> None:
        """
        Write data to the connection.
        For QUIC, this creates a new stream for each write operation.
        """
        if self._closed:
            raise QUICConnectionClosedError("Connection is closed")

        stream = await self.open_stream()
        try:
            await stream.write(data)
            await stream.close_write()
        except Exception:
            await stream.reset()
            raise

    async def read(self, n: int | None = -1) -> bytes:
        """
        Read data from the stream.

        Args:
            n: Maximum number of bytes to read. -1 means read all available.

        Returns:
            Data bytes read from the stream.

        Raises:
            QUICStreamClosedError: If stream is closed for reading.
            QUICStreamResetError: If stream was reset.
            QUICStreamTimeoutError: If read timeout occurs.

        """
        # It's here for interface compatibility but should not be used
        raise NotImplementedError(
            "Use streams for reading data from QUIC connections. "
            "Call accept_stream() or open_stream() instead."
        )

    # Utility and monitoring methods

    def get_stream_stats(self) -> dict[str, Any]:
        """Get stream statistics for monitoring."""
        return {
            "total_streams": len(self._streams),
            "outbound_streams": self._outbound_stream_count,
            "inbound_streams": self._inbound_stream_count,
            "max_streams": self.MAX_CONCURRENT_STREAMS,
            "stream_utilization": len(self._streams) / self.MAX_CONCURRENT_STREAMS,
            "stats": self._stats.copy(),
            "cache_size": len(
                self._stream_cache
            ),  # Include cache metrics for monitoring
        }

    def get_active_streams(self) -> list[QUICStream]:
        """Get list of active streams."""
        return [stream for stream in self._streams.values() if not stream.is_closed()]

    def get_streams_by_protocol(self, protocol: str) -> list[QUICStream]:
        """Get streams filtered by protocol."""
        return [
            stream
            for stream in self._streams.values()
            if hasattr(stream, "protocol")
            and stream.protocol == protocol
            and not stream.is_closed()
        ]

    def _update_stats(self) -> None:
        """Update connection statistics."""
        # Add any periodic stats updates here
        pass

    async def _cleanup_idle_streams(self) -> None:
        """Clean up idle streams that are no longer needed."""
        current_time = time.time()
        streams_to_cleanup = []

        for stream in self._streams.values():
            if stream.is_closed():
                # Check if stream has been closed for a while
                if hasattr(stream, "_timeline") and stream._timeline.closed_at:
                    if current_time - stream._timeline.closed_at > 60:  # 1 minute
                        streams_to_cleanup.append(stream.stream_id)

        for stream_id in streams_to_cleanup:
            self._remove_stream(int(stream_id))

    # String representation

    def __repr__(self) -> str:
        current_cid: str | None = (
            self._current_connection_id.hex() if self._current_connection_id else None
        )
        return (
            f"QUICConnection(peer={self._remote_peer_id}, "
            f"addr={self._remote_addr}, "
            f"initiator={self._is_initiator}, "
            f"verified={self._peer_verified}, "
            f"established={self._established}, "
            f"streams={len(self._streams)}, "
            f"current_cid={current_cid})"
        )

    def __str__(self) -> str:
        return f"QUICConnection({self._remote_peer_id})"



# Copied from https://github.com/ethereum/async-service

from typing import (
    NamedTuple,
)


class TaskStats(NamedTuple):
    total_count: int
    finished_count: int

    @property
    def pending_count(self) -> int:
        return self.total_count - self.finished_count


class Stats(NamedTuple):
    tasks: TaskStats



from collections.abc import (
    Sequence,
)


class BaseLibp2pError(Exception):
    pass


class ValidationError(BaseLibp2pError):
    """Raised when something does not pass a validation check."""


class ParseError(BaseLibp2pError):
    pass


class MultiError(BaseLibp2pError):
    r"""
    A combined error that wraps multiple exceptions into a single error object.
    This error is raised when multiple exceptions need to be reported together,
    typically in scenarios where parallel operations or multiple validations fail.

    Example\:
    ---------
        >>> from nanolib.exceptions import MultiError
        >>> errors = [
        ...     ValueError("Invalid input"),
        ...     TypeError("Wrong type"),
        ...     RuntimeError("Operation failed")
        ... ]
        >>> multi_error = MultiError(errors)
        >>> print(multi_error)
        Error 1: Invalid input
        Error 2: Wrong type
        Error 3: Operation failed

    Note\:
    ------
        The string representation of this error will number and list all contained
        errors sequentially, making it easier to identify individual issues in
        complex error scenarios.

    """

    def __init__(self, errors: Sequence[Exception]) -> None:
        super().__init__(errors)
        self.errors = errors  # Storing list of errors

    def __str__(self) -> str:
        return "\n".join(
            f"Error {i + 1}: {error}" for i, error in enumerate(self.errors)
        )



from collections.abc import (
    Awaitable,
    Callable,
)
import logging

import trio

from nanolib.abc import (
    IHost,
    INetStream,
)
from nanolib.network.stream.exceptions import (
    StreamError,
)
from nanolib.network.swarm import (
    Swarm,
)
from nanolib.peer.peerinfo import (
    info_from_p2p_addr,
)

from .constants import (
    MAX_READ_LEN,
)


async def connect_swarm(swarm_0: Swarm, swarm_1: Swarm) -> None:
    peer_id = swarm_1.get_peer_id()
    addrs = tuple(
        addr
        for transport in swarm_1.listeners.values()
        for addr in transport.get_addrs()
    )
    swarm_0.peerstore.add_addrs(peer_id, addrs, 10000)

    # Add retry logic for more robust connection
    max_retries = 3
    retry_delay = 0.2
    last_error = None

    for attempt in range(max_retries):
        try:
            await swarm_0.dial_peer(peer_id)

            # Verify connection is established in both directions
            if (
                swarm_0.get_peer_id() in swarm_1.connections
                and swarm_1.get_peer_id() in swarm_0.connections
            ):
                return

            # Connection partially established, wait a bit for it to complete
            await trio.sleep(0.1)

            if (
                swarm_0.get_peer_id() in swarm_1.connections
                and swarm_1.get_peer_id() in swarm_0.connections
            ):
                return

            logging.debug(
                "Swarm connection verification failed on attempt"
                + f" {attempt + 1}, retrying..."
            )

        except Exception as e:
            last_error = e
            logging.debug(f"Swarm connection attempt {attempt + 1} failed: {e}")
            await trio.sleep(retry_delay)

    # If we got here, all retries failed
    if last_error:
        raise RuntimeError(
            f"Failed to connect swarms after {max_retries} attempts"
        ) from last_error
    else:
        err_msg = (
            "Failed to establish bidirectional swarm connection"
            + f" after {max_retries} attempts"
        )
        raise RuntimeError(err_msg)


async def connect(node1: IHost, node2: IHost) -> None:
    """Connect node1 to node2."""
    addr = node2.get_addrs()[0]
    info = info_from_p2p_addr(addr)

    # Add retry logic for more robust connection with timeout
    max_retries = 3
    retry_delay = 0.2
    last_error = None

    for attempt in range(max_retries):
        try:
            # Use timeout for each connection attempt
            with trio.move_on_after(5):  # 5 second timeout
                await node1.connect(info)

            # Verify connection is established in both directions
            if (
                node2.get_id() in node1.get_network().connections
                and node1.get_id() in node2.get_network().connections
            ):
                return

            # Connection partially established, wait a bit for it to complete
            await trio.sleep(0.1)

            if (
                node2.get_id() in node1.get_network().connections
                and node1.get_id() in node2.get_network().connections
            ):
                return

            logging.debug(
                f"Connection verification failed on attempt {attempt + 1}, retrying..."
            )

        except Exception as e:
            last_error = e
            logging.debug(f"Connection attempt {attempt + 1} failed: {e}")
            await trio.sleep(retry_delay)

    # If we got here, all retries failed
    if last_error:
        raise RuntimeError(
            f"Failed to connect after {max_retries} attempts"
        ) from last_error
    else:
        err_msg = (
            f"Failed to establish bidirectional connection after {max_retries} attempts"
        )
        raise RuntimeError(err_msg)


def create_echo_stream_handler(
    ack_prefix: str,
) -> Callable[[INetStream], Awaitable[None]]:
    async def echo_stream_handler(stream: INetStream) -> None:
        while True:
            try:
                read_string = (await stream.read(MAX_READ_LEN)).decode()
            except StreamError:
                break

            resp = ack_prefix + read_string
            try:
                await stream.write(resp.encode())
            except StreamError:
                break

    return echo_stream_handler
